from typing import Sequence

from loguru import logger
import torch
import torch.nn as nn
import torch.nn.functional as F

from .update import BasicUpdateBlock
from .extractor import BasicEncoder_resconv, Basic_Context_Encoder_resconv
from .corr import get_corr_block
from .utils import coords_grid, upflow2
from .update import BasicUpdateBlock
from .xcit import XCiT

from ptlflow.utils.registry import register_model
from ptlflow.utils.utils import forward_interpolate_batch
from ..base_model.base_model import BaseModel

try:
    import alt_cuda_corr
except:
    alt_cuda_corr = None


def downflow(flow, mode="bilinear", factor=0.125):
    old_size = (flow.shape[2], flow.shape[3])
    new_size = (int(factor * flow.shape[2]), int(factor * flow.shape[3]))
    u_scale = new_size[1] / old_size[1]
    v_scale = new_size[0] / old_size[0]
    resized_flow = F.interpolate(
        flow, size=new_size, mode=mode, align_corners=True
    )  # b 2 h w
    resized_flow_split = torch.split(resized_flow, 1, dim=1)
    rescaled_flow = torch.cat(
        [u_scale * resized_flow_split[0], v_scale * resized_flow_split[1]], dim=1
    )

    return rescaled_flow


class CCMR(BaseModel):
    pretrained_checkpoints = {
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/ccmr-kitti-612444b9.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/ccmr-sintel-e1760f37.ckpt",
    }

    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 4,
        iters: Sequence[int] = [8, 10, 15],
        alternate_corr: bool = True,
        lookup_pyramid_levels: int = 2,
        lookup_radius: int = 4,
        model_type: str = "CCMR",
        cnet_norm: str = "group",
        fnet_norm: str = "group",
        num_scales: int = 3,
        **kwargs,
    ) -> None:
        super().__init__(output_stride=32, loss_fn=None, **kwargs)

        self.corr_levels = corr_levels
        self.corr_radius = corr_radius
        self.iters = iters
        self.alternate_corr = alternate_corr
        self.lookup_pyramid_levels = lookup_pyramid_levels
        self.lookup_radius = lookup_radius
        self.model_type = model_type
        self.cnet_norm = cnet_norm
        self.fnet_norm = fnet_norm
        self.num_scales = num_scales

        # Initiating the improved multi-scale feature encoders:
        self.fnet = BasicEncoder_resconv(
            output_dim=256, norm_fn=self.fnet_norm, model_type=self.model_type
        )
        self.cnet = Basic_Context_Encoder_resconv(
            output_dim=256, norm_fn=self.cnet_norm, model_type=self.model_type
        )
        self.update_block = BasicUpdateBlock(
            correlation_depth=(2 * self.lookup_radius + 1) ** 2
            * self.lookup_pyramid_levels,
            hidden_dim=128,
            scale=2,
            num_heads=8,
            depth=1,
            mlp_ratio=1,
            num_scales=self.num_scales,
        )

        # Initiate global context grouping modules
        self.xcit = nn.ModuleList(
            [XCiT(embed_dim=128, depth=1, num_heads=8, mlp_ratio=1, separate=False)]
        )
        for i in range(self.num_scales - 1):
            self.xcit.extend(
                [XCiT(embed_dim=128, depth=1, num_heads=8, mlp_ratio=1, separate=False)]
            )

        if self.alternate_corr and alt_cuda_corr is None:
            logger.warning(
                "!!! alt_cuda_corr is not compiled! The slower IterativeCorrBlock will be used instead !!!"
            )

    def freeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()

    def initialize_flow16(self, img):
        """Flow is represented as difference between two coordinate grids flow = coords1 - coords0"""
        N, C, H, W = img.shape
        coords0 = coords_grid(N, H // 16, W // 16, dtype=img.dtype, device=img.device)
        coords1 = coords_grid(N, H // 16, W // 16, dtype=img.dtype, device=img.device)

        # optical flow computed as difference: flow = coords1 - coords0
        return coords0, coords1

    def get_grid(self, img, scale):
        """Flow is represented as difference between two coordinate grids flow = coords1 - coords0"""
        N, C, H, W = img.shape
        coords0 = coords_grid(
            N, H // scale, W // scale, dtype=img.dtype, device=img.device
        )
        return coords0

    def upsample_flow(self, flow, mask, scale=8):
        """Upsample flow field [H/scale, W/scale, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, scale, scale, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(scale * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, scale * H, scale * W)

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        image1 = images[:, 0]
        image2 = images[:, 1]

        fnet_pyramid = self.fnet([image1, image2])
        cnet_pyramid = self.cnet(image1)

        coords0, coords1 = self.initialize_flow16(image1)

        flow_predictions = []

        if (
            inputs.get("prev_preds") is not None
            and inputs["prev_preds"].get("flow_small") is not None
        ):
            forward_flow = forward_interpolate_batch(inputs["prev_preds"]["flow_small"])
            coords1 = coords1 + forward_flow

        assert len(fnet_pyramid) == len(
            cnet_pyramid
        ), "fnet and cnet pyramid should have the same length."
        assert len(fnet_pyramid) == len(
            self.iters
        ), "pyramid levels and the length of GRU iteration lists should be the same."
        upsampling_offset = (
            self.num_scales - 1 if self.num_scales == 4 else self.num_scales
        )
        for index, (fmap1, fmap2) in enumerate(fnet_pyramid):
            corr_fn = get_corr_block(
                fmap1=fmap1,
                fmap2=fmap2,
                radius=self.lookup_radius,
                num_levels=self.lookup_pyramid_levels,
                alternate_corr=self.alternate_corr,
            )

            net, inp = torch.split(cnet_pyramid[index], [128, 128], dim=1)
            net = torch.tanh(net)
            inp = torch.relu(inp)
            global_context = self.xcit[index](inp)

            for itr in range(self.iters[index]):
                coords1 = coords1.detach()
                if index >= 1 and itr == 0:
                    flow = self.upsample_flow(
                        coords1 - coords0, up_mask, scale=2
                    )  # bug of MS-RAFT's flow upsampling fixed here.

                    coords0 = self.get_grid(image1, scale=16 / (2**index))

                    coords1 = coords0 + flow

                corr = corr_fn(coords1)
                flow = coords1 - coords0
                net, up_mask, delta_flow = self.update_block(
                    net, inp, corr, flow, global_context, level_index=index
                )

                # F(t+1) = F(t) + \Delta(t)
                coords1 = coords1 + delta_flow
                # upsample predictions
                flow_up = self.upsample_flow(coords1 - coords0, up_mask, scale=2)
                for i in range(upsampling_offset - index):
                    flow_up = upflow2(flow_up)

                flow_up = self.postprocess_predictions(
                    flow_up, image_resizer, is_flow=True
                )
                flow_predictions.append(flow_up)

        if self.training:
            outputs = {"flows": flow_up[:, None], "flow_preds": flow_predictions}
        else:
            outputs = {
                "flows": flow_up[:, None],
                "flow_small": downflow(flow_up, factor=0.0625),
            }

        return outputs


class CCMRPlus(CCMR):
    pretrained_checkpoints = {
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/ccmr+-kitti-c289d5e6.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/ccmr+-sintel-055b44ec.ckpt",
    }

    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 4,
        iters: Sequence[int] = [8, 10, 10, 10],
        alternate_corr: bool = True,
        lookup_pyramid_levels: int = 2,
        lookup_radius: int = 4,
        model_type: str = "CCMR+",
        cnet_norm: str = "group",
        fnet_norm: str = "group",
        num_scales: int = 4,
        **kwargs,
    ) -> None:
        super().__init__(
            corr_levels,
            corr_radius,
            iters,
            alternate_corr,
            lookup_pyramid_levels,
            lookup_radius,
            model_type,
            cnet_norm,
            fnet_norm,
            num_scales,
            **kwargs,
        )


@register_model
class ccmr(CCMR):
    pass


@register_model
class ccmr_p(CCMRPlus):
    pass



from argparse import ArgumentParser, Namespace

import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.registry import register_model
from .Networks.encoders import twins_svt_large, convnext_Xlarge_4x, convnext_base_2x
from .Networks.MOFNetStack.corr import CorrBlock, AlternateCorrBlock
from .utils import coords_grid
from .Networks.MOFNetStack.gma import Attention
from ..base_model.base_model import BaseModel


class VideoFlowMOF(BaseModel):
    pretrained_checkpoints = {
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/videoflow_mof-kitti-293b4f59.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/videoflow_mof-sintel-739e4d3a.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/videoflow_mof-things-e24551af.ckpt",
        "things_288960": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/videoflow_mof-things_288960noise-0615a42e.ckpt",
    }

    def __init__(
        self,
        corr_levels: int = 4,
        corr_radius: int = 4,
        cnet: str = "twins",
        fnet: str = "twins",
        gma: str = "GMA-SK2",
        pretrain: bool = True,
        corr_fn: str = "default",
        decoder_depth: int = 32,
        feat_dim: int = 256,
        Tfusion: str = "stack",
        down_ratio: int = 8,
        context_3D: bool = False,
        cost_heads_num: int = 1,
        **kwargs,
    ) -> None:
        super().__init__(loss_fn=None, output_stride=8, **kwargs)

        self.corr_levels = corr_levels
        self.corr_radius = corr_radius
        self.cnet = cnet
        self.fnet = fnet
        self.gma = gma
        self.pretrain = pretrain
        self.corr_fn = corr_fn
        self.decoder_depth = decoder_depth
        self.feat_dim = feat_dim
        self.Tfusion = Tfusion
        self.down_ratio = down_ratio
        self.context_3D = context_3D
        self.cost_heads_num = cost_heads_num

        self.hidden_dim = self.feat_dim // 2
        self.context_dim = self.feat_dim // 2

        # feature network, context network, and update block
        if cnet == "twins":
            print("[Using twins as context encoder]")
            self.cnet = twins_svt_large(pretrained=self.pretrain)
        elif cnet == "convnext_Xlarge_4x":
            print("[Using convnext_Xlarge_4x as context encoder]")
            self.cnet = convnext_Xlarge_4x(pretrained=self.pretrain)
        elif cnet == "convnext_base_2x":
            print("[Using convnext_base_2x as context encoder]")
            self.cnet = convnext_base_2x(pretrained=self.pretrain)

        if fnet == "twins":
            print("[Using twins as feature encoder]")
            self.fnet = twins_svt_large(pretrained=self.pretrain)
        elif fnet == "convnext_Xlarge_4x":
            print("[Using convnext_Xlarge_4x as feature encoder]")
            self.fnet = convnext_Xlarge_4x(pretrained=self.pretrain)
        elif fnet == "convnext_base_2x":
            print("[Using convnext_base_2x as feature encoder]")
            self.fnet = convnext_base_2x(pretrained=self.pretrain)

        hidden_dim_ratio = 256 // feat_dim

        if self.Tfusion == "stack":
            print("[Using stack.]")
            self.cost_heads_num = 1
            from .Networks.MOFNetStack.stack import (
                SKUpdateBlock6_Deep_nopoolres_AllDecoder2,
            )

            self.update_block = SKUpdateBlock6_Deep_nopoolres_AllDecoder2(
                feat_dim=feat_dim,
                down_ratio=down_ratio,
                corr_radius=corr_radius,
                corr_levels=corr_levels,
                cost_heads_num=cost_heads_num,
                hidden_dim=128 // hidden_dim_ratio,
            )

        print("[Using corr_fn {}]".format(self.corr_fn))

        self.att = Attention(
            dim=128 // hidden_dim_ratio,
            heads=1,
            max_pos_size=160,
            dim_head=128 // hidden_dim_ratio,
        )

        if self.context_3D:
            print("[Using 3D Conv on context feature.]")
            self.context_3D = nn.Sequential(
                nn.Conv3d(256, 256, 3, stride=1, padding=1),
                nn.GELU(),
                nn.Conv3d(256, 256, 3, stride=1, padding=1),
                nn.GELU(),
                nn.Conv3d(256, 256, 3, stride=1, padding=1),
                nn.GELU(),
            )

        self.has_showed_warning = False

    def initialize_flow(self, img, bs, down_ratio):
        """Flow is represented as difference between two coordinate grids flow = coords1 - coords0"""
        N, C, H, W = img.shape
        coords0 = coords_grid(
            bs, H // down_ratio, W // down_ratio, dtype=img.dtype, device=img.device
        )
        coords1 = coords_grid(
            bs, H // down_ratio, W // down_ratio, dtype=img.dtype, device=img.device
        )

        # optical flow computed as difference: flow = coords1 - coords0
        return coords0, coords1

    def upsample_flow(self, flow, mask):
        """Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 8, 8, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(8 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 8 * H, 8 * W)

    def upsample_flow_4x(self, flow, mask):
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 4, 4, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(4 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 4 * H, 4 * W)

    def upsample_flow_2x(self, flow, mask):
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 2, 2, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(2 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 2 * H, 2 * W)

    def forward(self, inputs):
        """Estimate optical flow between pair of frames"""
        down_ratio = self.down_ratio

        images = self._check_input_shape(inputs["images"].clone())
        N_orig = images.shape[1]

        num_left_reps = 0
        if images.shape[1] > 3 and "meta" in inputs and "image_paths" in inputs["meta"]:
            # Remove multiple repeated images to the left and right
            # VideoFlow only accepts at most one repetition on each side
            paths = [p[0] for p in inputs["meta"]["image_paths"]]
            num_left_reps = 0
            i = 1
            while i < len(paths) and paths[0] == paths[i]:
                num_left_reps += 1
                i += 1
            num_right_reps = 0
            i = len(paths) - 2
            while i >= 0 and paths[-1] == paths[i]:
                num_right_reps += 1
                i -= 1
            if num_left_reps > 0:
                images = images[:, num_left_reps - 1 :]
            if num_right_reps > 0:
                images = images[:, : images.shape[1] - num_right_reps]

        images, image_resizer = self.preprocess_images(
            images,
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )

        B, N, _, H, W = images.shape

        hdim = self.hidden_dim
        cdim = self.context_dim

        fmaps = self.fnet(images.reshape(B * N, 3, H, W)).reshape(
            B, N, -1, H // down_ratio, W // down_ratio
        )

        if self.corr_fn == "default":
            corr_fn = CorrBlock
        elif self.corr_fn == "efficient":
            corr_fn = AlternateCorrBlock

        forward_corr_fn = corr_fn(
            fmaps[:, 1 : N - 1, ...].reshape(
                B * (N - 2), -1, H // down_ratio, W // down_ratio
            ),
            fmaps[:, 2:N, ...].reshape(
                B * (N - 2), -1, H // down_ratio, W // down_ratio
            ),
            num_levels=self.corr_levels,
            radius=self.corr_radius,
        )
        backward_corr_fn = corr_fn(
            fmaps[:, 1 : N - 1, ...].reshape(
                B * (N - 2), -1, H // down_ratio, W // down_ratio
            ),
            fmaps[:, 0 : N - 2, ...].reshape(
                B * (N - 2), -1, H // down_ratio, W // down_ratio
            ),
            num_levels=self.corr_levels,
            radius=self.corr_radius,
        )

        cnet = self.cnet(images[:, 1 : N - 1, ...].reshape(B * (N - 2), 3, H, W))
        if self.context_3D:
            # print("!@!@@#!@#!@")
            cnet = cnet.reshape(B, N - 2, -1, H // 2, W // 2).permute(0, 2, 1, 3, 4)
            cnet = self.context_3D(cnet) + cnet
            # print(cnet.shape)
            cnet = cnet.permute(0, 2, 1, 3, 4).reshape(
                B * (N - 2), -1, H // down_ratio, W // down_ratio
            )
        net, inp = torch.split(cnet, [hdim, cdim], dim=1)
        net = torch.tanh(net)
        inp = torch.relu(inp)
        attention = self.att(inp)

        forward_coords1, forward_coords0 = self.initialize_flow(
            images[:, 0, ...], bs=B * (N - 2), down_ratio=down_ratio
        )
        backward_coords1, backward_coords0 = self.initialize_flow(
            images[:, 0, ...], bs=B * (N - 2), down_ratio=down_ratio
        )

        flow_predictions = []  # forward flows followed by backward flows

        motion_hidden_state = None

        for itr in range(self.decoder_depth):
            forward_coords1 = forward_coords1.detach()
            backward_coords1 = backward_coords1.detach()

            forward_corr = forward_corr_fn(forward_coords1)
            backward_corr = backward_corr_fn(backward_coords1)

            forward_flow = forward_coords1 - forward_coords0
            backward_flow = backward_coords1 - backward_coords0

            net, motion_hidden_state, up_mask, delta_flow = self.update_block(
                net,
                motion_hidden_state,
                inp,
                forward_corr,
                backward_corr,
                forward_flow,
                backward_flow,
                forward_coords0,
                attention,
                bs=B,
            )

            forward_up_mask, backward_up_mask = torch.split(
                up_mask, [down_ratio**2 * 9, down_ratio**2 * 9], dim=1
            )

            forward_coords1 = forward_coords1 + delta_flow[:, 0:2, ...]
            backward_coords1 = backward_coords1 + delta_flow[:, 2:4, ...]

            # upsample predictions
            if down_ratio == 4:
                forward_flow_up = self.upsample_flow_4x(
                    forward_coords1 - forward_coords0, forward_up_mask
                )
                backward_flow_up = self.upsample_flow_4x(
                    backward_coords1 - backward_coords0, backward_up_mask
                )
            elif down_ratio == 2:
                forward_flow_up = self.upsample_flow_2x(
                    forward_coords1 - forward_coords0, forward_up_mask
                )
                backward_flow_up = self.upsample_flow_2x(
                    backward_coords1 - backward_coords0, backward_up_mask
                )
            elif down_ratio == 8:
                forward_flow_up = self.upsample_flow(
                    forward_coords1 - forward_coords0, forward_up_mask
                )
                backward_flow_up = self.upsample_flow(
                    backward_coords1 - backward_coords0, backward_up_mask
                )

            pred_mid = (N_orig - 2) // 2
            if num_left_reps > 0:
                pred_mid -= num_left_reps - 1

            forward_flow_up = forward_flow_up.reshape(B, N - 2, 2, H, W)[
                :, pred_mid : pred_mid + 1
            ]
            backward_flow_up = backward_flow_up.reshape(B, N - 2, 2, H, W)[
                :, pred_mid : pred_mid + 1
            ]
            forward_flow_up = self.postprocess_predictions(
                forward_flow_up, image_resizer, is_flow=True
            )
            backward_flow_up = self.postprocess_predictions(
                backward_flow_up, image_resizer, is_flow=True
            )
            flow_predictions.append(
                torch.cat([forward_flow_up, backward_flow_up], dim=1)
            )

        if self.training:
            outputs = {
                "flows": forward_flow_up,
                "flows_bw": backward_flow_up,
                "flow_preds": flow_predictions,
            }
        else:
            outputs = {
                "flows": forward_flow_up,
                "flows_bw": backward_flow_up,
                "flow_small": forward_coords1 - forward_coords0,
                "flow_bw_small": backward_coords1 - backward_coords0,
            }

        return outputs

    def _check_input_shape(self, images):
        if images.shape[1] == 2:
            if not self.has_showed_warning:
                print(
                    "Warning: videoflow_mof requires inputs of at least 3 frames, but the current input has only 2. "
                    "The first frame will be replicated to run, which may decrease the prediction accuracy. "
                    "If using validate.py or test.py, add the arguments seqlen_3-seqpos_middle to the [val/test]_dataset arg."
                )
                self.has_showed_warning = True
            images = torch.cat([images[:, :1], images], 1)
        return images


@register_model
class videoflow_mof(VideoFlowMOF):
    pass



import pdb
import torch.nn as nn
import math
import torch
from torch.nn.parameter import Parameter
import torch.nn.functional as F
from torch.nn import Module
from torch.nn.modules.conv import _ConvNd
from torch.nn.modules.utils import _quadruple
from torch.autograd import Variable
from torch.nn import Conv2d


def conv4d(data, filters, bias=None, permute_filters=True, use_half=False):
    """
    This is done by stacking results of multiple 3D convolutions, and is very slow.
    Taken from https://github.com/ignacio-rocco/ncnet
    """
    b, c, h, w, d, t = data.size()

    data = data.permute(
        2, 0, 1, 3, 4, 5
    ).contiguous()  # permute to avoid making contiguous inside loop

    # Same permutation is done with filters, unless already provided with permutation
    if permute_filters:
        filters = filters.permute(
            2, 0, 1, 3, 4, 5
        ).contiguous()  # permute to avoid making contiguous inside loop

    c_out = filters.size(1)
    if use_half:
        output = Variable(
            torch.HalfTensor(h, b, c_out, w, d, t), requires_grad=data.requires_grad
        )
    else:
        output = Variable(
            torch.zeros(h, b, c_out, w, d, t), requires_grad=data.requires_grad
        )

    padding = filters.size(0) // 2
    if use_half:
        Z = Variable(torch.zeros(padding, b, c, w, d, t).half())
    else:
        Z = Variable(torch.zeros(padding, b, c, w, d, t))

    if data.is_cuda:
        Z = Z.cuda(data.get_device())
        output = output.cuda(data.get_device())

    data_padded = torch.cat((Z, data, Z), 0)

    for i in range(output.size(0)):  # loop on first feature dimension
        # convolve with center channel of filter (at position=padding)
        output[i, :, :, :, :, :] = F.conv3d(
            data_padded[i + padding, :, :, :, :, :],
            filters[padding, :, :, :, :, :],
            bias=bias,
            stride=1,
            padding=padding,
        )
        # convolve with upper/lower channels of filter (at postions [:padding] [padding+1:])
        for p in range(1, padding + 1):
            output[i, :, :, :, :, :] = output[i, :, :, :, :, :] + F.conv3d(
                data_padded[i + padding - p, :, :, :, :, :],
                filters[padding - p, :, :, :, :, :],
                bias=None,
                stride=1,
                padding=padding,
            )
            output[i, :, :, :, :, :] = output[i, :, :, :, :, :] + F.conv3d(
                data_padded[i + padding + p, :, :, :, :, :],
                filters[padding + p, :, :, :, :, :],
                bias=None,
                stride=1,
                padding=padding,
            )

    output = output.permute(1, 2, 0, 3, 4, 5).contiguous()
    return output


class Conv4d(_ConvNd):
    """Applies a 4D convolution over an input signal composed of several input
    planes.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        bias=True,
        pre_permuted_filters=True,
    ):
        # stride, dilation and groups !=1 functionality not tested
        stride = 1
        dilation = 1
        groups = 1
        # zero padding is added automatically in conv4d function to preserve tensor size
        padding = 0
        kernel_size = _quadruple(kernel_size)
        stride = _quadruple(stride)
        padding = _quadruple(padding)
        dilation = _quadruple(dilation)
        super(Conv4d, self).__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            False,
            _quadruple(0),
            groups,
            bias,
        )
        # weights will be sliced along one dimension during convolution loop
        # make the looping dimension to be the first one in the tensor,
        # so that we don't need to call contiguous() inside the loop
        self.pre_permuted_filters = pre_permuted_filters
        if self.pre_permuted_filters:
            self.weight.data = self.weight.data.permute(2, 0, 1, 3, 4, 5).contiguous()
        self.use_half = False

    #    self.isbias = bias
    #    if not self.isbias:
    #        self.bn = torch.nn.BatchNorm1d(out_channels)

    def forward(self, input):
        out = conv4d(
            input,
            self.weight,
            bias=self.bias,
            permute_filters=not self.pre_permuted_filters,
            use_half=self.use_half,
        )  # filters pre-permuted in constructor
        #    if not self.isbias:
        #        b,c,u,v,h,w = out.shape
        #        out = self.bn(out.view(b,c,-1)).view(b,c,u,v,h,w)
        return out


class fullConv4d(torch.nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        bias=True,
        pre_permuted_filters=True,
    ):
        super(fullConv4d, self).__init__()
        self.conv = Conv4d(
            in_channels,
            out_channels,
            kernel_size,
            bias=bias,
            pre_permuted_filters=pre_permuted_filters,
        )
        self.isbias = bias
        if not self.isbias:
            self.bn = torch.nn.BatchNorm1d(out_channels)

    def forward(self, input):
        out = self.conv(input)
        if not self.isbias:
            b, c, u, v, h, w = out.shape
            out = self.bn(out.view(b, c, -1)).view(b, c, u, v, h, w)
        return out


class butterfly4D(torch.nn.Module):
    """
    butterfly 4d
    """

    def __init__(self, fdima, fdimb, withbn=True, full=True, groups=1):
        super(butterfly4D, self).__init__()
        self.proj = nn.Sequential(
            projfeat4d(fdima, fdimb, 1, with_bn=withbn, groups=groups),
            nn.ReLU(inplace=True),
        )
        self.conva1 = sepConv4dBlock(
            fdimb, fdimb, with_bn=withbn, stride=(2, 1, 1), full=full, groups=groups
        )
        self.conva2 = sepConv4dBlock(
            fdimb, fdimb, with_bn=withbn, stride=(2, 1, 1), full=full, groups=groups
        )
        self.convb3 = sepConv4dBlock(
            fdimb, fdimb, with_bn=withbn, stride=(1, 1, 1), full=full, groups=groups
        )
        self.convb2 = sepConv4dBlock(
            fdimb, fdimb, with_bn=withbn, stride=(1, 1, 1), full=full, groups=groups
        )
        self.convb1 = sepConv4dBlock(
            fdimb, fdimb, with_bn=withbn, stride=(1, 1, 1), full=full, groups=groups
        )

    # @profile
    def forward(self, x):
        out = self.proj(x)
        b, c, u, v, h, w = out.shape  # 9x9

        out1 = self.conva1(out)  # 5x5, 3
        _, c1, u1, v1, h1, w1 = out1.shape

        out2 = self.conva2(out1)  # 3x3, 9
        _, c2, u2, v2, h2, w2 = out2.shape

        out2 = self.convb3(out2)  # 3x3, 9

        tout1 = F.interpolate(
            out2.view(b, c, u2, v2, -1),
            (u1, v1, h2 * w2),
            mode="trilinear",
            align_corners=False,
        ).view(
            b, c, u1, v1, h2, w2
        )  # 5x5
        tout1 = F.interpolate(
            tout1.view(b, c, -1, h2, w2),
            (u1 * v1, h1, w1),
            mode="trilinear",
            align_corners=False,
        ).view(
            b, c, u1, v1, h1, w1
        )  # 5x5
        out1 = tout1 + out1
        out1 = self.convb2(out1)

        tout = F.interpolate(
            out1.view(b, c, u1, v1, -1),
            (u, v, h1 * w1),
            mode="trilinear",
            align_corners=False,
        ).view(b, c, u, v, h1, w1)
        tout = F.interpolate(
            tout.view(b, c, -1, h1, w1),
            (u * v, h, w),
            mode="trilinear",
            align_corners=False,
        ).view(b, c, u, v, h, w)
        out = tout + out
        out = self.convb1(out)

        return out


class projfeat4d(torch.nn.Module):
    """
    Turn 3d projection into 2d projection
    """

    def __init__(self, in_planes, out_planes, stride, with_bn=True, groups=1):
        super(projfeat4d, self).__init__()
        self.with_bn = with_bn
        self.stride = stride
        self.conv1 = nn.Conv3d(
            in_planes,
            out_planes,
            1,
            (stride, stride, 1),
            padding=0,
            bias=not with_bn,
            groups=groups,
        )
        self.bn = nn.BatchNorm3d(out_planes)

    def forward(self, x):
        b, c, u, v, h, w = x.size()
        x = self.conv1(x.view(b, c, u, v, h * w))
        if self.with_bn:
            x = self.bn(x)
        _, c, u, v, _ = x.shape
        x = x.view(b, c, u, v, h, w)
        return x


class sepConv4d(torch.nn.Module):
    """
    Separable 4d convolution block as 2 3D convolutions
    """

    def __init__(
        self,
        in_planes,
        out_planes,
        stride=(1, 1, 1),
        with_bn=True,
        ksize=3,
        full=True,
        groups=1,
    ):
        super(sepConv4d, self).__init__()
        bias = not with_bn
        self.isproj = False
        self.stride = stride[0]
        expand = 1

        if with_bn:
            if in_planes != out_planes:
                self.isproj = True
                self.proj = nn.Sequential(
                    nn.Conv2d(
                        in_planes, out_planes, 1, bias=bias, padding=0, groups=groups
                    ),
                    nn.BatchNorm2d(out_planes),
                )
            if full:
                self.conv1 = nn.Sequential(
                    nn.Conv3d(
                        in_planes * expand,
                        in_planes,
                        (1, ksize, ksize),
                        stride=(1, self.stride, self.stride),
                        bias=bias,
                        padding=(0, ksize // 2, ksize // 2),
                        groups=groups,
                    ),
                    nn.BatchNorm3d(in_planes),
                )
            else:
                self.conv1 = nn.Sequential(
                    nn.Conv3d(
                        in_planes * expand,
                        in_planes,
                        (1, ksize, ksize),
                        stride=1,
                        bias=bias,
                        padding=(0, ksize // 2, ksize // 2),
                        groups=groups,
                    ),
                    nn.BatchNorm3d(in_planes),
                )
            self.conv2 = nn.Sequential(
                nn.Conv3d(
                    in_planes,
                    in_planes * expand,
                    (ksize, ksize, 1),
                    stride=(self.stride, self.stride, 1),
                    bias=bias,
                    padding=(ksize // 2, ksize // 2, 0),
                    groups=groups,
                ),
                nn.BatchNorm3d(in_planes * expand),
            )
        else:
            if in_planes != out_planes:
                self.isproj = True
                self.proj = nn.Conv2d(
                    in_planes, out_planes, 1, bias=bias, padding=0, groups=groups
                )
            if full:
                self.conv1 = nn.Conv3d(
                    in_planes * expand,
                    in_planes,
                    (1, ksize, ksize),
                    stride=(1, self.stride, self.stride),
                    bias=bias,
                    padding=(0, ksize // 2, ksize // 2),
                    groups=groups,
                )
            else:
                self.conv1 = nn.Conv3d(
                    in_planes * expand,
                    in_planes,
                    (1, ksize, ksize),
                    stride=1,
                    bias=bias,
                    padding=(0, ksize // 2, ksize // 2),
                    groups=groups,
                )
            self.conv2 = nn.Conv3d(
                in_planes,
                in_planes * expand,
                (ksize, ksize, 1),
                stride=(self.stride, self.stride, 1),
                bias=bias,
                padding=(ksize // 2, ksize // 2, 0),
                groups=groups,
            )
        self.relu = nn.ReLU(inplace=True)

    # @profile
    def forward(self, x):
        b, c, u, v, h, w = x.shape
        x = self.conv2(x.view(b, c, u, v, -1))  # WTA convolution over (u,v)
        b, c, u, v, _ = x.shape
        x = self.relu(x)
        x = self.conv1(x.view(b, c, -1, h, w))  # spatial convolution over (x,y)
        b, c, _, h, w = x.shape

        if self.isproj:
            x = self.proj(x.view(b, c, -1, w))
        x = x.view(b, -1, u, v, h, w)
        return x


class sepConv4dBlock(torch.nn.Module):
    """
    Separable 4d convolution block as 2 2D convolutions and a projection
    layer
    """

    def __init__(
        self, in_planes, out_planes, stride=(1, 1, 1), with_bn=True, full=True, groups=1
    ):
        super(sepConv4dBlock, self).__init__()
        if in_planes == out_planes and stride == (1, 1, 1):
            self.downsample = None
        else:
            if full:
                self.downsample = sepConv4d(
                    in_planes,
                    out_planes,
                    stride,
                    with_bn=with_bn,
                    ksize=1,
                    full=full,
                    groups=groups,
                )
            else:
                self.downsample = projfeat4d(
                    in_planes, out_planes, stride[0], with_bn=with_bn, groups=groups
                )
        self.conv1 = sepConv4d(
            in_planes, out_planes, stride, with_bn=with_bn, full=full, groups=groups
        )
        self.conv2 = sepConv4d(
            out_planes, out_planes, (1, 1, 1), with_bn=with_bn, full=full, groups=groups
        )
        self.relu1 = nn.ReLU(inplace=True)
        self.relu2 = nn.ReLU(inplace=True)

    # @profile
    def forward(self, x):
        out = self.relu1(self.conv1(x))
        if self.downsample:
            x = self.downsample(x)
        out = self.relu2(x + self.conv2(out))
        return out


##import torch.backends.cudnn as cudnn
##cudnn.benchmark = True
# import time
##im = torch.randn(9,64,9,160,224).cuda()
##net = torch.nn.Conv3d(64, 64, 3).cuda()
##net = Conv4d(1,1,3,bias=True,pre_permuted_filters=True).cuda()
##net = sepConv4dBlock(2,2,stride=(1,1,1)).cuda()
#
##im = torch.randn(1,16,9,9,96,320).cuda()
##net = sepConv4d(16,16,with_bn=False).cuda()
#
##im = torch.randn(1,16,81,96,320).cuda()
##net = torch.nn.Conv3d(16,16,(1,3,3),padding=(0,1,1)).cuda()
#
##im = torch.randn(1,16,9,9,96*320).cuda()
##net = torch.nn.Conv3d(16,16,(3,3,1),padding=(1,1,0)).cuda()
#
##im = torch.randn(10000,10,9,9).cuda()
##net = torch.nn.Conv2d(10,10,3,padding=1).cuda()
#
##im = torch.randn(81,16,96,320).cuda()
##net = torch.nn.Conv2d(16,16,3,padding=1).cuda()
# c=   int(16 *1)
# cp = int(16 *1)
# h=int(96  *4)
# w=int(320 *4)
# k=3
# im = torch.randn(1,c,h,w).cuda()
# net = torch.nn.Conv2d(c,cp,k,padding=k//2).cuda()
#
# im2 = torch.randn(cp,k*k*c).cuda()
# im1 = F.unfold(im, (k,k), padding=k//2)[0]
#
#
# net(im)
# net(im)
# torch.mm(im2,im1)
# torch.mm(im2,im1)
# torch.cuda.synchronize()
# beg = time.time()
# for i in range(100):
#    net(im)
#    #im1 = F.unfold(im, (k,k), padding=k//2)[0]
#    torch.mm(im2,im1)
# torch.cuda.synchronize()
# print('%f'%((time.time()-beg)*10.))



"""Functions taken and adapted from https://github.com/princeton-vl/RAFT/blob/master/core/utils/utils.py."""

# Original license below

# BSD 3-Clause License

# Copyright (c) 2020, princeton-vl
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import re

import numpy as np
from scipy import interpolate
import torch
import torch.nn.functional as F


class InputPadder:
    """Pads images such that dimensions are divisible by stride."""

    def __init__(
        self,
        dims,
        stride=8,
        two_side_pad=True,
        pad_mode="replicate",
        pad_value=0.0,
        size=None,
    ):
        self.pad_mode = pad_mode
        self.pad_value = pad_value
        ht, wd = dims[-2:]
        if size is None:
            pad_ht = (((ht // stride) + 1) * stride - ht) % stride
            pad_wd = (((wd // stride) + 1) * stride - wd) % stride
        else:
            pad_ht = size[0] - ht
            pad_wd = size[1] - wd
        if two_side_pad:
            self._pad = [
                pad_wd // 2,
                pad_wd - pad_wd // 2,
                pad_ht // 2,
                pad_ht - pad_ht // 2,
            ]
        else:
            self._pad = [pad_wd // 2, pad_wd - pad_wd // 2, 0, pad_ht]

    def pad(self, x):
        in_shape = x.shape
        if len(in_shape) > 4:
            x = x.view(-1, *in_shape[-3:])
        x = F.pad(x, self._pad, mode=self.pad_mode, value=self.pad_value)
        if len(in_shape) > 4:
            x = x.view(*in_shape[:-2], *x.shape[-2:])
        return x

    def unpad(self, x):
        ht, wd = x.shape[-2:]
        c = [self._pad[2], ht - self._pad[3], self._pad[0], wd - self._pad[1]]
        return x[..., c[0] : c[1], c[2] : c[3]]


def read_pfm(file_path: str) -> np.ndarray:
    """Read optical flow files in PFM format.

    Taken and adapted from https://github.com/princeton-vl/RAFT/blob/master/core/utils/frame_utils.py.

    Parameters
    ----------
    file_path : str
        Path to the optical flow file.

    Returns
    -------
    np.ndarray
        The optical flow in HC format.

    Raises
    ------
    ValueError
        If the file is not in valid PFM format.
    ValueError
        If the file header is corrupted.
    """
    with open(file_path, "rb") as f:
        color = None
        width = None
        height = None
        scale = None
        endian = None

        header = f.readline().rstrip()
        if header == b"PF":
            color = True
        elif header == b"Pf":
            color = False
        else:
            raise ValueError("Not a PFM file.")

        dim_match = re.match(rb"^(\d+)\s(\d+)\s$", f.readline())
        if dim_match:
            width, height = map(int, dim_match.groups())
        else:
            raise ValueError("Malformed PFM header.")

        scale = float(f.readline().rstrip())
        if scale < 0:  # little-endian
            endian = "<"
            scale = -scale
        else:
            endian = ">"  # big-endian

        data = np.fromfile(f, endian + "f")
        shape = (height, width, 3) if color else (height, width)

        data = np.reshape(data, shape)
        data = np.flipud(data)

    # Mark invalid pixels as NaN
    if color:
        mask = np.tile(data[:, :, 2:3], (1, 1, 2))
        flow = data[:, :, :2].astype(np.float32)
        flow[mask > 0.5] = float("nan")
    else:
        flow = data.astype(np.float32)
    return flow


def forward_interpolate(flow):
    flow = flow.detach().cpu().numpy()
    dx, dy = flow[0], flow[1]

    ht, wd = dx.shape
    x0, y0 = np.meshgrid(np.arange(wd), np.arange(ht))

    x1 = x0 + dx
    y1 = y0 + dy

    x1 = x1.reshape(-1)
    y1 = y1.reshape(-1)
    dx = dx.reshape(-1)
    dy = dy.reshape(-1)

    valid = (x1 > 0) & (x1 < wd) & (y1 > 0) & (y1 < ht)
    x1 = x1[valid]
    y1 = y1[valid]
    dx = dx[valid]
    dy = dy[valid]

    flow_x = interpolate.griddata(
        (x1, y1), dx, (x0, y0), method="nearest", fill_value=0
    )

    flow_y = interpolate.griddata(
        (x1, y1), dy, (x0, y0), method="nearest", fill_value=0
    )

    flow = np.stack([flow_x, flow_y], axis=0)
    return torch.from_numpy(flow).float()


def bilinear_sampler(img, coords, mode="bilinear", mask=False):
    """Wrapper for grid_sample, uses pixel coordinates"""
    H, W = img.shape[-2:]
    xgrid, ygrid = coords.split([1, 1], dim=-1)
    xgrid = 2 * xgrid / (W - 1) - 1
    ygrid = 2 * ygrid / (H - 1) - 1

    grid = torch.cat([xgrid, ygrid], dim=-1)
    img = F.grid_sample(img, grid, mode=mode, align_corners=True)

    if mask:
        mask = (xgrid > -1) & (ygrid > -1) & (xgrid < 1) & (ygrid < 1)
        return img, mask.to(dtype=coords.dtype)

    return img



from typing import Optional, Sequence

try:
    from spatial_correlation_sampler import SpatialCorrelationSampler
except ModuleNotFoundError:
    from ptlflow.utils.correlation import (
        IterSpatialCorrelationSampler as SpatialCorrelationSampler,
    )
import torch
import torch.nn as nn

from ptlflow.utils.registry import register_model
from .pwc_modules import conv, upsample2d_as, rescale_flow, initialize_msra
from .pwc_modules import WarpingLayer, FeatureExtractor
from .pwc_modules import FlowAndOccContextNetwork, FlowAndOccEstimatorDense
from .irr_modules import OccUpsampleNetwork, RefineFlow, RefineOcc
from ptlflow.models.base_model.base_model import BaseModel


class StarFlow(BaseModel):
    pretrained_checkpoints = {
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/starflow-things-d3966a93.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/starflow-sintel-3c06b844.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/starflow-kitti-34b9a3ed.ckpt",
    }

    def __init__(
        self,
        div_flow: float = 0.05,
        search_range: int = 4,
        output_level: int = 4,
        num_levels: int = 7,
        num_chs: Sequence[int] = (3, 16, 32, 64, 96, 128, 196),
        train_batch_size: Optional[int] = None,
        **kwargs,
    ):
        super(StarFlow, self).__init__(
            output_stride=64,
            loss_fn=None,
            **kwargs,
        )

        self.div_flow = div_flow
        self.search_range = search_range
        self.output_level = output_level
        self.num_levels = num_levels
        self.num_chs = num_chs
        self.train_batch_size = train_batch_size

        self.leakyRELU = nn.LeakyReLU(0.1, inplace=True)

        self.feature_pyramid_extractor = FeatureExtractor(self.num_chs)
        self.warping_layer = WarpingLayer()

        self.dim_corr = (self.search_range * 2 + 1) ** 2
        self.num_ch_in = self.dim_corr + 32 + 2 + 1

        self.flow_and_occ_estimators = FlowAndOccEstimatorDense(2 * self.num_ch_in)
        self.context_networks = FlowAndOccContextNetwork(
            2 * self.num_ch_in + 448 + 2 + 1
        )

        self.occ_shuffle_upsample = OccUpsampleNetwork(11, 1)

        self.conv_1x1 = nn.ModuleList(
            [
                conv(196, 32, kernel_size=1, stride=1, dilation=1),
                conv(128, 32, kernel_size=1, stride=1, dilation=1),
                conv(96, 32, kernel_size=1, stride=1, dilation=1),
                conv(64, 32, kernel_size=1, stride=1, dilation=1),
            ]
        )

        self.conv_1x1_1 = conv(16, 3, kernel_size=1, stride=1, dilation=1)

        self.conv_1x1_time = conv(
            2 * self.num_ch_in + 448,
            self.num_ch_in,
            kernel_size=1,
            stride=1,
            dilation=1,
        )

        self.refine_flow = RefineFlow(2 + 1 + 32)
        self.refine_occ = RefineOcc(1 + 32 + 32)

        self.corr = SpatialCorrelationSampler(
            kernel_size=1, patch_size=2 * self.search_range + 1, padding=0
        )

        initialize_msra(self.modules())

    def forward(self, inputs):
        images, image_resizer = self.preprocess_images(
            inputs["images"],
            bgr_add=0.0,
            bgr_mult=1.0,
            bgr_to_rgb=True,
            resize_mode="interpolation",
            interpolation_mode="bilinear",
            interpolation_align_corners=False,
        )
        list_imgs = [images[:, i] for i in range(images.shape[1])]

        _, _, height_im, width_im = list_imgs[0].size()

        # on the bottom level are original images
        list_pyramids = []  # indices : [time][level]
        for im in list_imgs:
            list_pyramids.append(self.feature_pyramid_extractor(im) + [im])

        # outputs
        output_dict = {}
        output_dict_eval = {}
        flows_f = []  # indices : [level][time]
        flows_b = []  # indices : [level][time]
        occs_f = []
        occs_b = []
        flows_coarse_f = []
        occs_coarse_f = []
        for l in range(len(list_pyramids[0])):
            flows_f.append([])
            flows_b.append([])
            occs_f.append([])
            occs_b.append([])
        for l in range(self.output_level + 1):
            flows_coarse_f.append([])
            occs_coarse_f.append([])

        # init
        (
            b_size,
            _,
            h_x1,
            w_x1,
        ) = list_pyramids[
            0
        ][0].size()
        init_dtype = list_pyramids[0][0].dtype
        init_device = list_pyramids[0][0].device
        flow_f = torch.zeros(
            b_size, 2, h_x1, w_x1, dtype=init_dtype, device=init_device
        )
        flow_b = torch.zeros(
            b_size, 2, h_x1, w_x1, dtype=init_dtype, device=init_device
        )
        occ_f = torch.zeros(b_size, 1, h_x1, w_x1, dtype=init_dtype, device=init_device)
        occ_b = torch.zeros(b_size, 1, h_x1, w_x1, dtype=init_dtype, device=init_device)
        previous_features = []

        for i in range(len(list_imgs) - 1):
            x1_pyramid, x2_pyramid = list_pyramids[i : i + 2]

            for l, (x1, x2) in enumerate(zip(x1_pyramid, x2_pyramid)):
                if l <= self.output_level:
                    if i == 0:
                        (
                            bs_,
                            _,
                            h_,
                            w_,
                        ) = list_pyramids[
                            0
                        ][l].size()
                        previous_features.append(
                            torch.zeros(
                                bs_,
                                self.num_ch_in,
                                h_,
                                w_,
                                dtype=init_dtype,
                                device=init_device,
                            )
                        )

                    # warping
                    if l == 0:
                        x2_warp = x2
                        x1_warp = x1
                    else:
                        flow_f = upsample2d_as(flow_f, x1, mode="bilinear")
                        flow_b = upsample2d_as(flow_b, x2, mode="bilinear")
                        occ_f = upsample2d_as(occ_f, x1, mode="bilinear")
                        occ_b = upsample2d_as(occ_b, x2, mode="bilinear")
                        x2_warp = self.warping_layer(
                            x2, flow_f, height_im, width_im, self.div_flow
                        )
                        x1_warp = self.warping_layer(
                            x1, flow_b, height_im, width_im, self.div_flow
                        )

                    # correlation
                    out_corr_f = self.corr(x1, x2_warp)
                    out_corr_f = out_corr_f.view(
                        out_corr_f.shape[0],
                        -1,
                        out_corr_f.shape[3],
                        out_corr_f.shape[4],
                    )
                    out_corr_f = out_corr_f / x1.shape[1]

                    out_corr_b = self.corr(x2, x1_warp)
                    out_corr_b = out_corr_b.view(
                        out_corr_b.shape[0],
                        -1,
                        out_corr_b.shape[3],
                        out_corr_b.shape[4],
                    )
                    out_corr_b = out_corr_b / x2.shape[1]

                    out_corr_relu_f = self.leakyRELU(out_corr_f)
                    out_corr_relu_b = self.leakyRELU(out_corr_b)

                    if l != self.output_level:
                        x1_1by1 = self.conv_1x1[l](x1)
                        x2_1by1 = self.conv_1x1[l](x2)
                    else:
                        x1_1by1 = x1
                        x2_1by1 = x2

                    if i > 0:  # temporal connection:
                        previous_features[l] = self.warping_layer(
                            previous_features[l],
                            flows_b[l][-1],
                            height_im,
                            width_im,
                            self.div_flow,
                        )

                    # Flow and occlusions estimation
                    flow_f = rescale_flow(
                        flow_f, self.div_flow, width_im, height_im, to_local=True
                    )
                    flow_b = rescale_flow(
                        flow_b, self.div_flow, width_im, height_im, to_local=True
                    )

                    features = torch.cat(
                        [previous_features[l], out_corr_relu_f, x1_1by1, flow_f, occ_f],
                        1,
                    )
                    features_b = torch.cat(
                        [
                            torch.zeros_like(previous_features[l]),
                            out_corr_relu_b,
                            x2_1by1,
                            flow_b,
                            occ_b,
                        ],
                        1,
                    )

                    x_intm_f, flow_res_f, occ_res_f = self.flow_and_occ_estimators(
                        features
                    )
                    flow_est_f = flow_f + flow_res_f
                    occ_est_f = occ_f + occ_res_f
                    with torch.no_grad():
                        x_intm_b, flow_res_b, occ_res_b = self.flow_and_occ_estimators(
                            features_b
                        )
                        flow_est_b = flow_b + flow_res_b
                        occ_est_b = occ_b + occ_res_b

                    # Context:
                    flow_cont_res_f, occ_cont_res_f = self.context_networks(
                        torch.cat([x_intm_f, flow_est_f, occ_est_f], dim=1)
                    )
                    flow_cont_f = flow_est_f + flow_cont_res_f
                    occ_cont_f = occ_est_f + occ_cont_res_f
                    with torch.no_grad():
                        flow_cont_res_b, occ_cont_res_b = self.context_networks(
                            torch.cat([x_intm_b, flow_est_b, occ_est_b], dim=1)
                        )
                        flow_cont_b = flow_est_b + flow_cont_res_b
                        occ_cont_b = occ_est_b + occ_cont_res_b

                    # refinement
                    img1_resize = upsample2d_as(list_imgs[i], flow_f, mode="bilinear")
                    img2_resize = upsample2d_as(
                        list_imgs[i + 1], flow_b, mode="bilinear"
                    )
                    flow_cont_f = rescale_flow(
                        flow_cont_f,
                        self.div_flow,
                        width_im,
                        height_im,
                        to_local=False,
                    )
                    flow_cont_b = rescale_flow(
                        flow_cont_b,
                        self.div_flow,
                        width_im,
                        height_im,
                        to_local=False,
                    )
                    img2_warp = self.warping_layer(
                        img2_resize,
                        flow_cont_f,
                        height_im,
                        width_im,
                        self.div_flow,
                    )
                    img1_warp = self.warping_layer(
                        img1_resize,
                        flow_cont_b,
                        height_im,
                        width_im,
                        self.div_flow,
                    )

                    # flow refine
                    flow_f = self.refine_flow(
                        flow_cont_f.detach(), img1_resize - img2_warp, x1_1by1
                    )
                    flow_b = self.refine_flow(
                        flow_cont_b.detach(), img2_resize - img1_warp, x2_1by1
                    )

                    flow_f = rescale_flow(
                        flow_f, self.div_flow, width_im, height_im, to_local=False
                    )
                    flow_b = rescale_flow(
                        flow_b, self.div_flow, width_im, height_im, to_local=False
                    )

                    # occ refine
                    x2_1by1_warp = self.warping_layer(
                        x2_1by1, flow_f, height_im, width_im, self.div_flow
                    )
                    x1_1by1_warp = self.warping_layer(
                        x1_1by1, flow_b, height_im, width_im, self.div_flow
                    )

                    occ_f = self.refine_occ(
                        occ_cont_f.detach(), x1_1by1, x1_1by1 - x2_1by1_warp
                    )
                    occ_b = self.refine_occ(
                        occ_cont_b.detach(), x2_1by1, x2_1by1 - x1_1by1_warp
                    )

                    # save features for temporal connection:
                    previous_features[l] = self.conv_1x1_time(x_intm_f)
                    flows_f[l].append(flow_f)
                    occs_f[l].append(occ_f)
                    flows_b[l].append(flow_b)
                    occs_b[l].append(occ_b)
                    flows_coarse_f[l].append(flow_cont_f)
                    occs_coarse_f[l].append(occ_cont_f)
                    # flows.append([flow_cont_f, flow_cont_b, flow_f, flow_b])
                    # occs.append([occ_cont_f, occ_cont_b, occ_f, occ_b])

                else:
                    flow_f = upsample2d_as(flow_f, x1, mode="bilinear")
                    flow_b = upsample2d_as(flow_b, x2, mode="bilinear")
                    flows_f[l].append(flow_f)
                    flows_b[l].append(flow_b)
                    # flows.append([flow_f, flow_b])

                    x2_warp = self.warping_layer(
                        x2, flow_f, height_im, width_im, self.div_flow
                    )
                    x1_warp = self.warping_layer(
                        x1, flow_b, height_im, width_im, self.div_flow
                    )
                    flow_b_warp = self.warping_layer(
                        flow_b, flow_f, height_im, width_im, self.div_flow
                    )
                    flow_f_warp = self.warping_layer(
                        flow_f, flow_b, height_im, width_im, self.div_flow
                    )

                    if l != self.num_levels - 1:
                        x1_in = self.conv_1x1_1(x1)
                        x2_in = self.conv_1x1_1(x2)
                        x1_w_in = self.conv_1x1_1(x1_warp)
                        x2_w_in = self.conv_1x1_1(x2_warp)
                    else:
                        x1_in = x1
                        x2_in = x2
                        x1_w_in = x1_warp
                        x2_w_in = x2_warp

                    occ_f = self.occ_shuffle_upsample(
                        occ_f, torch.cat([x1_in, x2_w_in, flow_f, flow_b_warp], dim=1)
                    )
                    occ_b = self.occ_shuffle_upsample(
                        occ_b, torch.cat([x2_in, x1_w_in, flow_b, flow_f_warp], dim=1)
                    )

                    occs_f[l].append(occ_f)
                    occs_b[l].append(occ_b)
                    # occs.append([occ_f, occ_b])

            flow_f = torch.zeros(
                b_size, 2, h_x1, w_x1, dtype=init_dtype, device=init_device
            )
            flow_b = torch.zeros(
                b_size, 2, h_x1, w_x1, dtype=init_dtype, device=init_device
            )
            occ_f = torch.zeros(
                b_size, 1, h_x1, w_x1, dtype=init_dtype, device=init_device
            )
            occ_b = torch.zeros(
                b_size, 1, h_x1, w_x1, dtype=init_dtype, device=init_device
            )

        flow_f_up = torch.stack(
            [
                upsample2d_as(f, list_imgs[0], mode="bilinear") / self.div_flow
                for f in flows_f[-1]
            ],
            dim=1,
        )
        flow_f_up = self.postprocess_predictions(flow_f_up, image_resizer, is_flow=True)
        flow_b_up = torch.stack(
            [
                upsample2d_as(f, list_imgs[0], mode="bilinear") / self.div_flow
                for f in flows_b[-1]
            ],
            dim=1,
        )
        flow_b_up = self.postprocess_predictions(flow_b_up, image_resizer, is_flow=True)
        occs_f_up = torch.stack(
            [
                upsample2d_as(torch.sigmoid(o), list_imgs[0], mode="bilinear")
                for o in occs_f[-1]
            ],
            dim=1,
        )
        occs_f_up = self.postprocess_predictions(
            occs_f_up, image_resizer, is_flow=False
        )
        occs_b_up = torch.stack(
            [
                upsample2d_as(torch.sigmoid(o), list_imgs[0], mode="bilinear")
                for o in occs_b[-1]
            ],
            dim=1,
        )
        occs_b_up = self.postprocess_predictions(
            occs_b_up, image_resizer, is_flow=False
        )

        outputs = {}

        if self.training:
            raise NotImplementedError("Training is still not implemented for StarFlow.")
        else:
            outputs["flows"] = flow_f_up
            outputs["occs"] = occs_f_up
            outputs["flows_b"] = flow_b_up
            outputs["occs_b"] = occs_b_up

        return outputs


@register_model
class starflow(StarFlow):
    pass



# =============================================================================
# Copyright 2025 Henrique Morimitsu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import math
from typing import Optional

from loguru import logger
import torch
import torch.nn as nn
import torch.nn.functional as F

from ..base_model.base_model import BaseModel
from .corr import CorrBlock, AlternateCorrBlock
from .pwc_modules import rescale_flow, upsample2d_as
from .cgu_bidir_dual_encoder import CGUBidirDualEncoder
from .update import UpdateBlock
from .utils import (
    compute_pyramid_levels,
    get_activation,
    get_norm,
)
from ptlflow.utils.utils import forward_interpolate_batch
from ptlflow.utils.registry import register_model, trainable, ptlflow_trained

try:
    import alt_cuda_corr
except:
    alt_cuda_corr = None


class SequenceLoss(nn.Module):
    def __init__(self, loss: str, max_flow: float, gamma: float):
        super().__init__()
        self.loss = loss
        self.max_flow = max_flow
        self.gamma = gamma

    def forward(self, outputs, inputs):
        """Loss function defined over sequence of flow predictions"""

        flow_preds = outputs["flow_preds"]
        flow_gt = inputs["flows"][:, 0]
        valid = inputs["valids"][:, 0]

        n_predictions = len(flow_preds)
        flow_loss = 0.0

        # exclude invalid pixels and extremely large diplacements
        mag = torch.sum(flow_gt**2, dim=1, keepdim=True).sqrt()
        valid = (valid >= 0.5) & (mag < self.max_flow)

        for i in range(n_predictions):
            pred = flow_preds[i]
            if (
                pred.shape[-2] != flow_gt.shape[-2]
                or pred.shape[-1] != flow_gt.shape[-1]
            ):
                pred = F.interpolate(
                    pred, size=flow_gt.shape[-2:], mode="bilinear", align_corners=True
                )
            i_weight = self.gamma ** (n_predictions - i - 1)

            if self.loss == "l1" or outputs["nf_preds"][i] is None:
                diff = pred - flow_gt
                i_loss = (diff).abs()
                valid_loss = valid * i_loss
                flow_loss += i_weight * valid_loss.mean()
            elif self.loss == "laplace":
                loss_i = outputs["nf_preds"][i]
                final_mask = (
                    (~torch.isnan(loss_i.detach()))
                    & (~torch.isinf(loss_i.detach()))
                    & valid
                )
                flow_loss += i_weight * ((final_mask * loss_i).sum() / final_mask.sum())

        return flow_loss


class DPFlow(BaseModel):
    pretrained_checkpoints = {
        "chairs": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/dpflow-chairs-f94e717a.ckpt",
        "kitti": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/dpflow-kitti-4e97eac6.ckpt",
        "sintel": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/dpflow-sintel-b44b072c.ckpt",
        "spring": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/dpflow-spring-69bac7fa.ckpt",
        "things": "https://github.com/hmorimitsu/ptlflow/releases/download/weights1/dpflow-things-2012b5d6.ckpt",
    }

    def __init__(
        self,
        pyramid_levels: Optional[int] = None,
        iters_per_level: int = 4,
        detach_flow: bool = True,
        use_norm_affine: bool = False,
        group_norm_num_groups: int = 8,
        corr_mode: str = "allpairs",  # "allpairs" or "local"
        corr_levels: int = 1,
        corr_range: int = 4,
        activation_function: str = "orig",  # "orig", "relu", "gelu", "silu", or "mish"
        enc_network: str = "cgu_bidir_dual",  # "cgu", "cgu_bidir", "cgu_bidir_dual", "cgu_dual", "next_bidir_dual", "swin"
        enc_norm_type: str = "group",  # "none", "group", "layer", or "batch"
        enc_depth: int = 4,
        enc_mlp_ratio: float = 2.0,
        enc_mlp_in_kernel_size: int = 1,
        enc_mlp_out_kernel_size: int = 1,
        enc_hidden_chs: list[int] = (64, 96, 128),
        enc_num_out_stages: int = 1,
        enc_out_1x1_chs: str = "384",
        dec_gru_norm_type: str = "layer",  # "none", "group", "layer", or "batch"
        dec_gru_iters: int = 1,
        dec_gru_depth: int = 4,
        dec_gru_mlp_ratio: float = 2.0,
        dec_gru_mlp_in_kernel_size: int = 1,
        dec_gru_mlp_out_kernel_size: int = 1,
        dec_net_chs: int = 128,
        dec_inp_chs: int = 128,
        dec_motion_chs: int = 128,
        dec_flow_kernel_size: int = 7,
        dec_flow_head_chs: int = 256,
        dec_motenc_corr_hidden_chs: int = 256,
        dec_motenc_corr_out_chs: int = 192,
        dec_motenc_flow_hidden_chs: int = 128,
        dec_motenc_flow_out_chs: int = 64,
        use_upsample_mask: bool = True,
        upmask_gradient_scale: float = 1.0,
        cgu_mlp_dw_kernel_size: int = 7,
        cgu_fusion_gate_activation: str = "gelu",  # "linear", "sigmoid", "relu", "gelu", "silu", or "mish"
        cgu_mlp_use_dw_conv: bool = True,
        cgu_mlp_activation_function: str = "gelu",  # "linear", "sigmoid", "relu", "gelu", "silu", or "mish"
        cgu_layer_scale_init_value: float = 0.01,
        loss: str = "laplace",  # "l1" or "laplace"
        gamma: float = 0.8,
        max_flow: float = 400.0,
        use_var: bool = True,
        var_min: float = 0.0,
        var_max: float = 10.0,
        **kwargs,
    ):
        if pyramid_levels is not None:
            assert pyramid_levels > 2, "Only --model.pyramid_levels >= 3 is supported."
            output_stride = int(2 ** (pyramid_levels + 2))
            if enc_network == "swin_bidir_dual":
                output_stride *= 2
        else:
            logger.info(
                f"DPFlow: --model.pyramid_levels is not set, the number of pyramid levels will be inferred from the input size."
            )
            output_stride = None
            self.extra_output_stride = 1 if enc_network == "swin_bidir_dual" else 0

        super(DPFlow, self).__init__(
            loss_fn=SequenceLoss(loss=loss, max_flow=max_flow, gamma=gamma),
            output_stride=output_stride,
            **kwargs,
        )

        self.pyramid_levels = pyramid_levels
        self.iters_per_level = iters_per_level
        self.corr_mode = corr_mode
        self.corr_range = corr_range
        self.corr_levels = corr_levels
        self.detach_flow = detach_flow
        self.loss = loss
        self.use_var = use_var
        self.var_min = var_min
        self.var_max = var_max

        activation_function = get_activation(activation_function)

        enc_out_1x1_chs = (
            float(enc_out_1x1_chs)
            if (isinstance(enc_out_1x1_chs, str) and "." in enc_out_1x1_chs)
            else int(enc_out_1x1_chs)
        )

        if isinstance(enc_out_1x1_chs, float):
            out_1x1_factor = enc_out_1x1_chs
            out_1x1_abs_chs = int(enc_out_1x1_chs * enc_hidden_chs[-1])
        else:
            out_1x1_factor = None
            out_1x1_abs_chs = enc_out_1x1_chs

        self.max_feat_chs = max(
            enc_hidden_chs[-1],
            out_1x1_abs_chs,
        )

        net_chs = dec_net_chs
        inp_chs = dec_inp_chs
        if net_chs is None or inp_chs is None:
            base_chs = out_1x1_abs_chs
            if base_chs < 1:
                base_chs = enc_hidden_chs[-1]

            base_chs = base_chs // 3 * 2

            if net_chs is None and inp_chs is None:
                net_chs = inp_chs = base_chs // 2
            elif net_chs is None and inp_chs is not None:
                net_chs = base_chs - inp_chs
            elif net_chs is not None and inp_chs is None:
                inp_chs = base_chs - net_chs
        net_chs_fixed = net_chs
        inp_chs_fixed = inp_chs

        enc_norm_layer = get_norm(
            enc_norm_type,
            affine=use_norm_affine,
            num_groups=group_norm_num_groups,
        )
        self.fnet = CGUBidirDualEncoder(
            pyramid_levels=pyramid_levels,
            hidden_chs=enc_hidden_chs,
            out_1x1_abs_chs=out_1x1_abs_chs,
            out_1x1_factor=out_1x1_factor,
            num_out_stages=enc_num_out_stages,
            activation_function=activation_function,
            norm_layer=enc_norm_layer,
            depth=enc_depth,
            mlp_ratio=enc_mlp_ratio,
            mlp_use_dw_conv=cgu_mlp_use_dw_conv,
            mlp_dw_kernel_size=cgu_mlp_dw_kernel_size,
            mlp_in_kernel_size=enc_mlp_in_kernel_size,
            mlp_out_kernel_size=enc_mlp_out_kernel_size,
            cgu_layer_scale_init_value=cgu_layer_scale_init_value,
        )

        self.dim_corr = (corr_range * 2 + 1) ** 2 * corr_levels

        dec_gru_norm_layer = get_norm(
            dec_gru_norm_type,
            affine=use_norm_affine,
            num_groups=group_norm_num_groups,
        )
        self.update_block = UpdateBlock(
            dec_motenc_corr_hidden_chs=dec_motenc_corr_hidden_chs,
            dec_motenc_corr_out_chs=dec_motenc_corr_out_chs,
            dec_motenc_flow_hidden_chs=dec_motenc_flow_hidden_chs,
            dec_motenc_flow_out_chs=dec_motenc_flow_out_chs,
            corr_levels=corr_levels,
            corr_range=corr_range,
            dec_flow_kernel_size=dec_flow_kernel_size,
            dec_motion_chs=dec_motion_chs,
            activation_function=activation_function,
            net_chs_fixed=net_chs_fixed,
            inp_chs_fixed=inp_chs_fixed,
            dec_gru_norm_layer=dec_gru_norm_layer,
            dec_gru_depth=dec_gru_depth,
            dec_gru_iters=dec_gru_iters,
            dec_gru_mlp_ratio=dec_gru_mlp_ratio,
            cgu_mlp_use_dw_conv=cgu_mlp_use_dw_conv,
            cgu_mlp_dw_kernel_size=cgu_mlp_dw_kernel_size,
            dec_gru_mlp_in_kernel_size=dec_gru_mlp_in_kernel_size,
            dec_gru_mlp_out_kernel_size=dec_gru_mlp_out_kernel_size,
            cgu_layer_scale_init_value=cgu_layer_scale_init_value,
            dec_flow_head_chs=dec_flow_head_chs,
            loss=loss,
            use_upsample_mask=use_upsample_mask,
            upmask_gradient_scale=upmask_gradient_scale,
        )

        act = nn.ReLU if activation_function is None else activation_function
        self.input_act = act(inplace=True)

        self.current_output_stride = output_stride

        self.has_shown_input_message = False
        self.has_shown_altcuda_message = False

    def coords_grid(self, batch, ht, wd):
        coords = torch.meshgrid(
            torch.arange(ht, dtype=self.dtype, device=self.device),
            torch.arange(wd, dtype=self.dtype, device=self.device),
            indexing="ij",
        )
        coords = torch.stack(coords[::-1], dim=0).to(dtype=self.dtype)
        return coords[None].repeat(batch, 1, 1, 1)

    def upsample_flow(self, flow, mask, factor, ch=2):
        """Upsample flow field [H/f, W/f, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, factor, factor, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, ch, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, ch, factor * H, factor * W)

    def _show_input_message(self, images):
        pyr_levels = compute_pyramid_levels(images)
        recommended_pyr_levels = pyr_levels  # 3 for 1K, 4 for 2K, etc.

        logger.info(
            f"DPFlow: Using {self.pyramid_levels} pyramid levels and {self.iters_per_level} iterations per level."
        )
        logger.info(
            f"DPFlow: Processing inputs of resolution {images.shape[-1]} x {images.shape[-2]}"
        )
        logger.info(f"DPFlow: Correlation mode: {self.corr_mode}")

        if recommended_pyr_levels != self.pyramid_levels:
            logger.info(
                "DPFlow: For this input size, you may get better results by setting --pyramid_levels {}",
                recommended_pyr_levels,
            )

    def _show_altcuda_message(self):
        if self.corr_mode == "local" and alt_cuda_corr is None:
            logger.warning(
                f"DPFlow: You are running with --corr_mode local, but alt_cuda_corr is not installed. Please install alt_cuda_corr to increase the speed."
            )

    def forward(self, inputs):
        try:
            return self.forward_flow(inputs)
        except torch.OutOfMemoryError:
            if self.corr_mode == "allpairs":
                logger.warning(
                    "DPFlow: CUDA out of memory error with input size {}. DPFlow will set --model.corr_mode to 'local' and re-attempt inference. This decreases memory consumption, but it is also slower.",
                    list(inputs["images"].shape[-2:]),
                )
                self.corr_mode = "local"
                try:
                    return self.forward_flow(inputs)
                except torch.OutOfMemoryError:
                    logger.error(
                        "DPFlow: CUDA out of memory error even after setting --model.corr_mode to 'local'. DPFlow cannot process this input size: {} on this device.",
                        list(inputs["images"].shape[-2:]),
                    )
            else:
                logger.error(
                    "DPFlow: CUDA out of memory error even with --model.corr_mode set to 'local'. DPFlow cannot process this input size: {} on this device.",
                    list(inputs["images"].shape[-2:]),
                )

    def forward_flow(self, inputs):
        if self.corr_mode == "local" and not self.has_shown_altcuda_message:
            self._show_altcuda_message()
            self.has_shown_altcuda_message = True

        if self.pyramid_levels is not None and not self.has_shown_input_message:
            self._show_input_message(inputs["images"])
            self.has_shown_input_message = True

        if self.pyramid_levels is None:
            pyr_levels = compute_pyramid_levels(inputs["images"])
            output_stride = 2 ** (pyr_levels + 2 + self.extra_output_stride)

            if output_stride != self.current_output_stride:
                logger.info(
                    "DPFlow: Detected change in input size. The number of pyramid levels will change to {}, corresponding to output stride {}.",
                    pyr_levels,
                    output_stride,
                )
                self.current_output_stride = output_stride
        else:
            pyr_levels = self.pyramid_levels
            output_stride = self.output_stride

        images, image_resizer = self.preprocess_images(
            inputs["images"],
            stride=output_stride,
            bgr_add=-0.5,
            bgr_mult=2.0,
            bgr_to_rgb=True,
            resize_mode="pad",
            pad_mode="replicate",
            pad_two_side=True,
        )
        image1 = images[:, 0]
        image2 = images[:, 1]

        flow_init = None
        if (
            inputs.get("prev_preds") is not None
            and inputs["prev_preds"].get("flow_small") is not None
        ):
            flow_init = inputs["prev_preds"]["flow_small"]

        flow_predictions, flow_small, flow_up, info_predictions = self.predict(
            image1,
            image2,
            pyr_levels=pyr_levels,
            image_resizer=image_resizer,
            flow_init=flow_init,
        )

        nf_predictions = []
        if self.training and self.loss == "laplace":
            # exlude invalid pixels and extremely large diplacements
            for i in range(len(info_predictions)):
                if not self.use_var:
                    var_max = var_min = 0
                else:
                    var_max = self.var_max
                    var_min = self.var_min

                if info_predictions[i] is None:
                    nf_predictions.append(None)
                else:
                    raw_b = info_predictions[i][:, 2:]
                    log_b = torch.zeros_like(raw_b)
                    weight = info_predictions[i][:, :2]
                    # Large b Component
                    log_b[:, 0] = torch.clamp(raw_b[:, 0], min=0, max=var_max)
                    # Small b Component
                    log_b[:, 1] = torch.clamp(raw_b[:, 1], min=var_min, max=0)
                    # term2: [N, 2, m, H, W]
                    term2 = (
                        (inputs["flows"][:, 0] - flow_predictions[i]).abs().unsqueeze(2)
                    ) * (torch.exp(-log_b).unsqueeze(1))
                    # term1: [N, m, H, W]
                    term1 = weight - math.log(2) - log_b
                    nf_loss = torch.logsumexp(
                        weight, dim=1, keepdim=True
                    ) - torch.logsumexp(term1.unsqueeze(1) - term2, dim=2)
                    nf_predictions.append(nf_loss)

        outputs = {"flows": flow_up[:, None], "flow_small": flow_small}

        if self.training:
            outputs["flow_preds"] = flow_predictions
            outputs["nf_preds"] = nf_predictions

        return outputs

    def predict(self, x1_raw, x2_raw, pyr_levels, image_resizer, flow_init=None):
        b, _, height_im, width_im = x1_raw.size()

        x1_pyramid, x2_pyramid = self.fnet(x1_raw, x2_raw, pyr_levels=pyr_levels)

        # outputs
        flows = []
        infos = []

        # init
        (
            b_size,
            _,
            h_x1,
            w_x1,
        ) = x1_pyramid[0].size()
        init_device = x1_pyramid[0].device

        if flow_init is not None:
            flow = flow_init
            flow = rescale_flow(
                flow,
                x1_pyramid[0].shape[-1],
                x1_pyramid[0].shape[-2],
                to_local=False,
            )
            flow = upsample2d_as(flow, x1_pyramid[0], mode="bilinear")
            flow = forward_interpolate_batch(flow)
        else:
            flow = torch.zeros(
                b_size, 2, h_x1, w_x1, dtype=self.dtype, device=init_device
            )

        net = None
        for l, (x1, x2) in enumerate(zip(x1_pyramid, x2_pyramid)):
            # Split feature channels into matching (x) and context (c)
            xh = x1.shape[1]
            ch = xh // 3
            x1, cn1 = torch.split(x1, [xh - ch, ch], dim=1)
            x2, cn2 = torch.split(x2, [xh - ch, ch], dim=1)
            halfch = ch // 2
            i1, n1 = torch.split(cn1, [ch - halfch, halfch], dim=1)
            i2, n2 = torch.split(cn2, [ch - halfch, halfch], dim=1)
            inp = torch.cat([i1, i2], 1)
            inp = self.input_act(inp)
            net_tmp = torch.cat([n1, n2], 1)

            coords0 = self.coords_grid(x1.shape[0], x1.shape[2], x1.shape[3])

            if self.corr_mode == "allpairs":
                corr_fn = CorrBlock(x1, x2, self.corr_levels, self.corr_range)
            else:
                corr_fn = AlternateCorrBlock(x1, x2, self.corr_levels, self.corr_range)

            if l > 0:
                flow = rescale_flow(flow, x1.shape[-1], x1.shape[-2], to_local=False)
                flow = upsample2d_as(flow, x1, mode="bilinear")

            net = torch.tanh(net_tmp)

            for it in range(self.iters_per_level):
                if self.detach_flow:
                    flow = flow.detach()

                # correlation
                out_corr = corr_fn(coords0 + flow)

                flow_res, net, mask = self.update_block(net, inp, out_corr, flow)

                info = None
                if self.loss == "laplace":
                    info = flow_res[:, 2:]
                    flow_res = flow_res[:, :2]

                flow = flow + flow_res

                if self.training or (
                    l == len(x1_pyramid) - 1 and it == self.iters_per_level - 1
                ):
                    out_flow = rescale_flow(flow, width_im, height_im, to_local=False)
                    if mask is not None:
                        out_flow = self.upsample_flow(out_flow, mask, factor=8)
                    out_flow = upsample2d_as(out_flow, x1_raw, mode="bilinear")
                    out_flow = self.postprocess_predictions(
                        out_flow, image_resizer, is_flow=True
                    )
                    flows.append(out_flow)

                    out_info = None
                    if info is not None:
                        if mask is not None:
                            out_info = self.upsample_flow(info, mask, factor=8, ch=4)
                        out_info = upsample2d_as(out_info, x1_raw, mode="bilinear")
                        out_info = self.postprocess_predictions(
                            out_info, image_resizer, is_flow=False
                        )
                    infos.append(out_info)

        return flows, flow, out_flow, infos


@register_model
@trainable
@ptlflow_trained
class dpflow(DPFlow):
    pass



from setuptools import setup
from torch.utils.cpp_extension import BuildExtension, CUDAExtension


setup(
    name="correlation",
    ext_modules=[
        CUDAExtension(
            "alt_cuda_corr",
            sources=["correlation.cpp", "correlation_kernel.cu"],
            extra_compile_args={"cxx": [], "nvcc": ["-O3"]},
        ),
    ],
    cmdclass={"build_ext": BuildExtension},
)



import torch
import torch.nn as nn
import timm
from einops import rearrange
from torch import nn
from timm.layers import to_2tuple


class PatchEmbed(nn.Module):
    """Image to Patch Embedding"""

    def __init__(self, patch_size=4, in_chans=3, embed_dim=768):
        super().__init__()
        patch_size = to_2tuple(patch_size)
        self.patch_size = patch_size
        self.proj = nn.Conv2d(
            in_chans, embed_dim, kernel_size=patch_size, stride=patch_size
        )
        self.norm = nn.LayerNorm(embed_dim)

    def forward(self, x):
        # x: B, T, C, H, W
        B, T, _, H, W = x.shape
        x = rearrange(x, "b t c h w -> (b t) c h w")
        x = self.proj(x)
        x = rearrange(x, "(b t) c h w -> b (t h w) c", b=B, t=T)
        x = self.norm(x)
        out_size = ((H * T) // self.patch_size[0], W // self.patch_size[1])

        return x, out_size


class Twins_CSC(nn.Module):
    def __init__(self, pretrained_ckpt=None):
        super().__init__()
        self.svt = timm.create_model("twins_svt_large", pretrained=False)
        self.svt.patch_embeds = nn.ModuleList()
        embed_dims = [128, 256, 512, 1024]

        self.svt.patch_embeds.append(PatchEmbed(4, 3, embed_dims[0]))  # 4倍下采样
        self.svt.patch_embeds.append(
            PatchEmbed(2, embed_dims[0], embed_dims[1])
        )  # 2倍下采样
        self.svt.patch_embeds.append(
            PatchEmbed(2, embed_dims[1], embed_dims[2])
        )  # 2倍下采样
        self.svt.patch_embeds.append(
            PatchEmbed(2, embed_dims[2], embed_dims[3])
        )  # 2倍下采样

        if pretrained_ckpt is not None:
            self.svt.load_state_dict(torch.load(pretrained_ckpt), strict=True)

        del self.svt.head
        del self.svt.blocks[2]
        del self.svt.blocks[2]
        del self.svt.pos_block[2]
        del self.svt.pos_block[2]
        del self.svt.patch_embeds[2]
        del self.svt.patch_embeds[2]

    def forward(self, x):
        layer = 2
        # if input is list, combine batch dimension
        B, T, C, H, W = x.shape
        ratios = [4, 2]

        for i, (embed, drop, blocks, pos_blk) in enumerate(
            zip(
                self.svt.patch_embeds,
                self.svt.pos_drops,
                self.svt.blocks,
                self.svt.pos_block,
            )
        ):

            x, size = embed(x)  # x: (B, T*h*w, C)
            x = drop(x)
            for j, blk in enumerate(blocks):
                x = blk(x, size)
                if j == 0:
                    x = pos_blk(x, size)

            # self.svt.depths实质为2
            H, W = H // ratios[i], W // ratios[i]
            x = rearrange(x, "b (t h w) c -> b t c h w", t=T, h=H, w=W)

            # x = x.reshape(B, *size, -1).permute(0, 3, 1, 2).contiguous()

            if i == layer - 1:
                break

        return x



import torch
import torch.nn as nn
from torch import nn, einsum
from einops import rearrange
import torch.nn.functional as F

from timm.layers import Mlp, DropPath, to_2tuple
from timm.models.vision_transformer import Attention as MyAttention


class ResidualBlock(nn.Module):
    def __init__(self, in_planes, planes, norm_fn="group", stride=1):
        super(ResidualBlock, self).__init__()

        self.conv1 = nn.Conv2d(
            in_planes, planes, kernel_size=3, padding=1, stride=stride
        )
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, padding=1)
        self.relu = nn.ReLU(inplace=True)

        num_groups = planes // 8

        if norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
            self.norm2 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
            if not stride == 1:
                self.norm3 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)

        elif norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(planes)
            self.norm2 = nn.BatchNorm2d(planes)
            if not stride == 1:
                self.norm3 = nn.BatchNorm2d(planes)

        elif norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(planes)
            self.norm2 = nn.InstanceNorm2d(planes)
            if not stride == 1:
                self.norm3 = nn.InstanceNorm2d(planes)

        elif norm_fn == "none":
            self.norm1 = nn.Sequential()
            self.norm2 = nn.Sequential()
            if not stride == 1:
                self.norm3 = nn.Sequential()

        if stride == 1:
            self.downsample = None

        else:
            self.downsample = nn.Sequential(
                nn.Conv2d(in_planes, planes, kernel_size=1, stride=stride), self.norm3
            )

    def forward(self, x):
        y = x
        y = self.relu(self.norm1(self.conv1(y)))
        y = self.relu(self.norm2(self.conv2(y)))

        if self.downsample is not None:
            x = self.downsample(x)

        return self.relu(x + y)


class BottleneckBlock(nn.Module):
    def __init__(self, in_planes, planes, norm_fn="group", stride=1):
        super(BottleneckBlock, self).__init__()

        self.conv1 = nn.Conv2d(in_planes, planes // 4, kernel_size=1, padding=0)
        self.conv2 = nn.Conv2d(
            planes // 4, planes // 4, kernel_size=3, padding=1, stride=stride
        )
        self.conv3 = nn.Conv2d(planes // 4, planes, kernel_size=1, padding=0)
        self.relu = nn.ReLU(inplace=True)

        num_groups = planes // 8

        if norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=num_groups, num_channels=planes // 4)
            self.norm2 = nn.GroupNorm(num_groups=num_groups, num_channels=planes // 4)
            self.norm3 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
            if not stride == 1:
                self.norm4 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)

        elif norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(planes // 4)
            self.norm2 = nn.BatchNorm2d(planes // 4)
            self.norm3 = nn.BatchNorm2d(planes)
            if not stride == 1:
                self.norm4 = nn.BatchNorm2d(planes)

        elif norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(planes // 4)
            self.norm2 = nn.InstanceNorm2d(planes // 4)
            self.norm3 = nn.InstanceNorm2d(planes)
            if not stride == 1:
                self.norm4 = nn.InstanceNorm2d(planes)

        elif norm_fn == "none":
            self.norm1 = nn.Sequential()
            self.norm2 = nn.Sequential()
            self.norm3 = nn.Sequential()
            if not stride == 1:
                self.norm4 = nn.Sequential()

        if stride == 1:
            self.downsample = None

        else:
            self.downsample = nn.Sequential(
                nn.Conv2d(in_planes, planes, kernel_size=1, stride=stride), self.norm4
            )

    def forward(self, x):
        y = x
        y = self.relu(self.norm1(self.conv1(y)))
        y = self.relu(self.norm2(self.conv2(y)))
        y = self.relu(self.norm3(self.conv3(y)))

        if self.downsample is not None:
            x = self.downsample(x)

        return self.relu(x + y)


class BasicEncoder(nn.Module):
    def __init__(self, output_dim=128, norm_fn="batch", dropout=0.0):
        super(BasicEncoder, self).__init__()
        self.norm_fn = norm_fn

        if self.norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=8, num_channels=64)

        elif self.norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(64)

        elif self.norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(64)

        elif self.norm_fn == "none":
            self.norm1 = nn.Sequential()

        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3)
        self.relu1 = nn.ReLU(inplace=True)

        self.in_planes = 64
        self.layer1 = self._make_layer(64, stride=1)
        self.layer2 = self._make_layer(96, stride=2)
        self.layer3 = self._make_layer(128, stride=2)

        # output convolution
        self.conv2 = nn.Conv2d(128, output_dim, kernel_size=1)

        self.dropout = None
        if dropout > 0:
            self.dropout = nn.Dropout2d(p=dropout)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.InstanceNorm2d, nn.GroupNorm)):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def _make_layer(self, dim, stride=1):
        layer1 = ResidualBlock(self.in_planes, dim, self.norm_fn, stride=stride)
        layer2 = ResidualBlock(dim, dim, self.norm_fn, stride=1)
        layers = (layer1, layer2)

        self.in_planes = dim
        return nn.Sequential(*layers)

    def forward(self, x):
        # if input is list, combine batch dimension
        is_list = isinstance(x, tuple) or isinstance(x, list)
        if is_list:
            len_list = len(x)
            batch_dim = x[0].shape[0]
            x = torch.cat(x, dim=0)

        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu1(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)

        x = self.conv2(x)

        if self.training and self.dropout is not None:
            x = self.dropout(x)

        if is_list:
            x = torch.split(x, [batch_dim for _ in range(len_list)], dim=0)

        return x


class ResidualBlock3D(nn.Module):
    def __init__(
        self, in_planes, planes, norm_fn="group", stride=1, drop_path=0.0, T=3
    ):
        super().__init__()

        self.conv1 = nn.Conv2d(
            in_planes, planes, kernel_size=3, padding=1, stride=stride
        )
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, padding=1)
        self.relu = nn.ReLU(inplace=True)

        num_groups = planes // 8

        if norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
            self.norm2 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)
            if not stride == 1:
                self.norm3 = nn.GroupNorm(num_groups=num_groups, num_channels=planes)

        elif norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(planes)
            self.norm2 = nn.BatchNorm2d(planes)
            if not stride == 1:
                self.norm3 = nn.BatchNorm2d(planes)

        elif norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(planes)
            self.norm2 = nn.InstanceNorm2d(planes)
            if not stride == 1:
                self.norm3 = nn.InstanceNorm2d(planes)

        elif norm_fn == "none":
            self.norm1 = nn.Sequential()
            self.norm2 = nn.Sequential()
            if not stride == 1:
                self.norm3 = nn.Sequential()

        if stride == 1:
            self.downsample = None

        else:
            self.downsample = nn.Sequential(
                nn.Conv2d(in_planes, planes, kernel_size=1, stride=stride), self.norm3
            )

        # temporal
        self.drop_path0 = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm0 = nn.LayerNorm(in_planes)
        self.temporal_attn = MyAttention(in_planes, 4, qkv_bias=False)
        self.temporal_fc = nn.Linear(in_planes, in_planes)
        self.T = T

        # init temporal blocks
        for name, m in self.named_modules():
            if "temporal_conv" in name:
                nn.init.dirac_(m.weight.data)  # initialized to be identity
                nn.init.zeros_(m.bias.data)
            if "temporal_fc" in name:
                nn.init.constant_(m.weight, 0)
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        T = self.T

        y = x
        BT, _, H, W = x.shape
        # print(BT, T)
        B = int(BT // T)
        # temporal
        y = rearrange(y, "(b t) c h w -> (b h w) t c", b=B, t=T, h=H, w=W)
        y = y + self.temporal_fc(self.drop_path0(self.temporal_attn(self.norm0(y))))
        y = rearrange(y, "(b h w) t c -> (b t) c h w", b=B, t=T, h=H, w=W)
        # spatial
        y = self.relu(self.norm1(self.conv1(y)))
        y = self.relu(self.norm2(self.conv2(y)))

        if self.downsample is not None:
            x = self.downsample(x)

        return self.relu(x + y)


class Res18_3D(nn.Module):
    def __init__(self, args, norm_fn="batch"):
        super().__init__()
        dropout = 0.0
        output_dim = 256
        self.norm_fn = norm_fn

        if self.norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=8, num_channels=64)

        elif self.norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(64)

        elif self.norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(64)

        elif self.norm_fn == "none":
            self.norm1 = nn.Sequential()

        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3)
        self.relu1 = nn.ReLU(inplace=True)

        self.in_planes = 64
        drop_rate = [x.item() for x in torch.linspace(0, 0.1, 6)]
        self.layer1 = self._make_layer(
            64, stride=1, drop_path0=drop_rate[0], drop_path1=drop_rate[1], T=args.T
        )
        self.layer2 = self._make_layer(
            96, stride=2, drop_path0=drop_rate[2], drop_path1=drop_rate[3], T=args.T
        )
        self.layer3 = self._make_layer(
            128, stride=2, drop_path0=drop_rate[4], drop_path1=drop_rate[5], T=args.T
        )

        # output convolution
        self.conv2 = nn.Conv2d(128, output_dim, kernel_size=1)

        self.dropout = None
        if dropout > 0:
            self.dropout = nn.Dropout2d(p=dropout)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.InstanceNorm2d, nn.GroupNorm)):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

        # init temporal blocks
        for name, m in self.named_modules():
            if "temporal_conv" in name:
                nn.init.dirac_(m.weight.data)  # initialized to be identity
                nn.init.zeros_(m.bias.data)
            if "temporal_fc" in name:
                nn.init.constant_(m.weight, 0)
                nn.init.constant_(m.bias, 0)

    def _make_layer(self, dim, stride=1, drop_path0=0.0, drop_path1=0.0, T=3):
        layer1 = ResidualBlock3D(
            self.in_planes, dim, self.norm_fn, stride=stride, drop_path=drop_path0, T=T
        )
        layer2 = ResidualBlock3D(
            dim, dim, self.norm_fn, stride=1, drop_path=drop_path1, T=T
        )
        layers = (layer1, layer2)

        self.in_planes = dim
        return nn.Sequential(*layers)

    def forward(self, x):
        B, T = x.shape[:2]
        x = rearrange(x, "b t c h w -> (b t) c h w")

        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu1(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)

        x = self.conv2(x)

        if self.training and self.dropout is not None:
            x = self.dropout(x)

        x = rearrange(x, "(b t) c h w -> b t c h w", b=B, t=T)

        return torch.split(x, 1, dim=1)


class MFBasicEncoder(nn.Module):
    def __init__(self, pretrained=True, args=None, norm_fn=None):
        output_dim = 256
        dropout = 0.0
        super().__init__()
        self.norm_fn = norm_fn

        if self.norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=8, num_channels=64)

        elif self.norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(64)

        elif self.norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(64)

        elif self.norm_fn == "none":
            self.norm1 = nn.Sequential()

        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3)
        self.relu1 = nn.ReLU(inplace=True)

        self.in_planes = 64
        self.layer1 = self._make_layer(64, stride=1)
        self.layer2 = self._make_layer(96, stride=2)
        self.layer3 = self._make_layer(128, stride=2)

        # output convolution
        self.conv2 = nn.Conv2d(128, output_dim, kernel_size=1)

        self.dropout = None
        if dropout > 0:
            self.dropout = nn.Dropout2d(p=dropout)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.InstanceNorm2d, nn.GroupNorm)):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def _make_layer(self, dim, stride=1):
        layer1 = ResidualBlock(self.in_planes, dim, self.norm_fn, stride=stride)
        layer2 = ResidualBlock(dim, dim, self.norm_fn, stride=1)
        layers = (layer1, layer2)

        self.in_planes = dim
        return nn.Sequential(*layers)

    def forward(self, x):
        # if input is list, combine batch dimension
        B, T = x.shape[:2]
        x = rearrange(x, "b t c h w -> (b t) c h w")

        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu1(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)

        x = self.conv2(x)

        if self.training and self.dropout is not None:
            x = self.dropout(x)

        x = rearrange(x, "(b t) c h w -> b t c h w", b=B, t=T)
        return torch.split(x, 1, dim=1)


class AttnBasicEncoder(nn.Module):
    def __init__(self, output_dim=128, norm_fn="batch", dropout=0.0):
        super().__init__()
        self.norm_fn = norm_fn

        if self.norm_fn == "group":
            self.norm1 = nn.GroupNorm(num_groups=8, num_channels=64)

        elif self.norm_fn == "batch":
            self.norm1 = nn.BatchNorm2d(64)

        elif self.norm_fn == "instance":
            self.norm1 = nn.InstanceNorm2d(64)

        elif self.norm_fn == "none":
            self.norm1 = nn.Sequential()

        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3)
        self.relu1 = nn.ReLU(inplace=True)

        self.in_planes = 64
        self.layer1 = self._make_layer(64, stride=1)
        self.layer2 = self._make_layer(96, stride=2)
        self.layer3 = self._make_layer(128, stride=2)

        # output convolution
        self.conv2 = nn.Conv2d(128, output_dim, kernel_size=1)

        self.dropout = None
        if dropout > 0:
            self.dropout = nn.Dropout2d(p=dropout)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.InstanceNorm2d, nn.GroupNorm)):
                if m.weight is not None:
                    nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def _make_layer(self, dim, stride=1):
        layer1 = ResidualBlock(self.in_planes, dim, self.norm_fn, stride=stride)
        layer2 = ResidualBlock(dim, dim, self.norm_fn, stride=1)
        layers = (layer1, layer2)

        self.in_planes = dim
        return nn.Sequential(*layers)

    def forward(self, x):
        # if input is list, combine batch dimension
        is_list = isinstance(x, tuple) or isinstance(x, list)
        if is_list:
            batch_dim = x[0].shape[0]
            x = torch.cat(x, dim=0)

        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu1(x)

        x = self.layer1(x)
        x = self.layer2(x)

        feat = rearrange(x[:batch_dim], "b (h d) x y -> b h x y d", h=1)
        self_attn = einsum("b h x y d, b h u v d -> b h x y u v", feat, feat)
        self_attn = rearrange(self_attn, "b h x y u v -> b h (x y) (u v)").softmax(
            dim=-1
        )

        x = self.layer3(x)
        x = self.conv2(x)

        if self.training and self.dropout is not None:
            x = self.dropout(x)

        if is_list:
            x1, x2 = torch.split(x, [batch_dim, batch_dim], dim=0)
            return x1, x2, self_attn

        return x



import torch
import torch.nn as nn
import torch.nn.functional as F

from ptlflow.utils.utils import forward_interpolate_batch
from ...utils import coords_grid, bilinear_sampler
from .attention import (
    MultiHeadAttention,
    LinearPositionEmbeddingSine,
    ExpPositionEmbeddingSine,
)

from timm.layers import DropPath

from .gru import BasicUpdateBlock, GMAUpdateBlock
from .gma import Attention
from .sk import SKUpdateBlock6_Deep_nopoolres_AllDecoder


def initialize_flow(img):
    """Flow is represented as difference between two means flow = mean1 - mean0"""
    N, C, H, W = img.shape
    mean = coords_grid(N, H, W, dtype=img.dtype, device=img.device)
    mean_init = coords_grid(N, H, W, dtype=img.dtype, device=img.device)

    # optical flow computed as difference: flow = mean1 - mean0
    return mean, mean_init


class CrossAttentionLayer(nn.Module):
    def __init__(
        self,
        qk_dim,
        v_dim,
        query_token_dim,
        tgt_token_dim,
        flow_or_pe,
        num_heads=8,
        attn_drop=0.0,
        proj_drop=0.0,
        drop_path=0.0,
        dropout=0.0,
        pe="linear",
        no_sc=False,
    ):
        super(CrossAttentionLayer, self).__init__()

        head_dim = qk_dim // num_heads
        self.scale = head_dim**-0.5
        self.query_token_dim = query_token_dim
        self.pe = pe

        self.norm1 = nn.LayerNorm(query_token_dim)
        self.norm2 = nn.LayerNorm(query_token_dim)
        self.multi_head_attn = MultiHeadAttention(qk_dim, num_heads)
        self.q, self.k, self.v = (
            nn.Linear(query_token_dim, qk_dim, bias=True),
            nn.Linear(tgt_token_dim, qk_dim, bias=True),
            nn.Linear(tgt_token_dim, v_dim, bias=True),
        )

        self.proj = nn.Linear(v_dim, query_token_dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

        self.ffn = nn.Sequential(
            nn.Linear(query_token_dim, query_token_dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(query_token_dim, query_token_dim),
            nn.Dropout(dropout),
        )
        self.flow_or_pe = flow_or_pe
        print("[Decoder flow_or_pe setting is: {}]".format(self.flow_or_pe))
        self.no_sc = no_sc
        if self.no_sc:
            print("[No short cut in cost decoding]")
        self.dim = qk_dim

    def forward(self, query, key, value, memory, query_coord, patch_size, size_h3w3):
        """
        query_coord [B, 2, H1, W1]
        """
        B, _, H1, W1 = query_coord.shape

        if key is None and value is None:
            key = self.k(memory)
            value = self.v(memory)

        # [B, 2, H1, W1] -> [BH1W1, 1, 2]
        query_coord = query_coord.contiguous()
        query_coord = (
            query_coord.view(B, 2, -1)
            .permute(0, 2, 1)[:, :, None, :]
            .contiguous()
            .view(B * H1 * W1, 1, 2)
        )
        if self.pe == "linear":
            query_coord_enc = LinearPositionEmbeddingSine(query_coord, dim=self.dim)
        elif self.pe == "exp":
            query_coord_enc = ExpPositionEmbeddingSine(query_coord, dim=self.dim)
        elif self.pe == "norm_linear":
            query_coord[:, :, 0:1] = query_coord[:, :, 0:1] / W1
            query_coord[:, :, 1:2] = query_coord[:, :, 1:2] / H1
            query_coord_enc = LinearPositionEmbeddingSine(
                query_coord, dim=self.dim, NORMALIZE_FACOR=2
            )

        short_cut = query
        if query is not None:
            query = self.norm1(query)

        if self.flow_or_pe == "and":
            q = self.q(query + query_coord_enc)
        elif self.flow_or_pe == "pe":
            q = self.q(query_coord_enc)
        elif self.flow_or_pe == "flow":
            q = self.q(query)
        else:
            print("[Wrong setting of flow_or_pe]")
            exit()
        k, v = key, value

        x = self.multi_head_attn(q, k, v)

        x = self.proj(x)
        # x = self.proj(torch.cat([x, short_cut],dim=2))
        if short_cut is not None and not self.no_sc:
            # print("short cut")
            x = short_cut + self.proj_drop(x)

        x = x + self.drop_path(self.ffn(self.norm2(x)))

        return x, k, v


class MemoryDecoderLayer(nn.Module):
    def __init__(
        self,
        query_latent_dim,
        cost_latent_dim,
        patch_size,
        flow_or_pe,
        dropout,
        pe,
        no_sc,
    ):
        super(MemoryDecoderLayer, self).__init__()
        self.query_latent_dim = query_latent_dim
        self.patch_size = patch_size  # for converting coords into H2', W2' space

        query_token_dim, tgt_token_dim = query_latent_dim, cost_latent_dim
        qk_dim, v_dim = query_token_dim, query_token_dim

        self.cross_attend = CrossAttentionLayer(
            qk_dim,
            v_dim,
            query_token_dim,
            tgt_token_dim,
            flow_or_pe=flow_or_pe,
            dropout=dropout,
            pe=pe,
            no_sc=no_sc,
        )

    def forward(self, query, key, value, memory, coords1, size, size_h3w3):
        """
        x:      [B*H1*W1, 1, C]
        memory: [B*H1*W1, H2'*W2', C]
        coords1 [B, 2, H2, W2]
        size: B, C, H1, W1
        1. Note that here coords0 and coords1 are in H2, W2 space.
           Should first convert it into H2', W2' space.
        2. We assume the upper-left point to be [0, 0], instead of letting center of upper-left patch to be [0, 0]
        """
        x_global, k, v = self.cross_attend(
            query, key, value, memory, coords1, self.patch_size, size_h3w3
        )
        B, C, H1, W1 = size
        C = self.query_latent_dim
        x_global = x_global.view(B, H1, W1, C).permute(0, 3, 1, 2)
        return x_global, k, v


class MemoryDecoder(nn.Module):
    def __init__(
        self,
        gma,
        use_patch,
        detach_local,
        use_rpe,
        r_16,
        quater_refine,
        fix_pe,
        gt_r,
        query_num,
        no_border,
        W_offset,
        H_offset,
        query_latent_dim,
        cost_latent_input_dim,
        cost_heads_num,
        encoder_latent_dim,
        decoder_depth,
        cost_latent_dim,
        patch_size,
        flow_or_pe,
        dropout,
        pe,
        no_sc,
    ):
        super(MemoryDecoder, self).__init__()
        self.gma = gma
        self.use_patch = use_patch
        self.detach_local = detach_local
        self.use_rpe = use_rpe
        self.r_16 = r_16
        self.quater_refine = quater_refine
        self.fix_pe = fix_pe
        self.gt_r = gt_r
        self.query_num = query_num
        self.no_border = no_border
        self.W_offset = W_offset
        self.H_offset = H_offset
        dim = self.dim = query_latent_dim

        if use_patch:
            print("[Using cost patch as local cost]")
            self.flow_token_encoder = nn.Conv2d(
                cost_latent_input_dim + 64, query_latent_dim, 1, 1
            )
        else:
            self.flow_token_encoder = nn.Sequential(
                nn.Conv2d(81 * cost_heads_num, dim, 1, 1),
                nn.GELU(),
                nn.Conv2d(dim, dim, 1, 1),
            )

        if fix_pe:
            print("[fix_pe: regress 8*8 block]")
            self.pretrain_head = nn.Sequential(
                nn.Conv2d(dim, dim * 2, 1, 1),
                nn.GELU(),
                nn.Conv2d(dim * 2, dim * 2, 1, 1),
                nn.GELU(),
                nn.Conv2d(dim * 2, 64, 1, 1),
            )
        elif gt_r > 0:
            print("[Using larger cost as gt, radius is {}]".format(gt_r))
            self.pretrain_head = nn.Sequential(
                nn.Conv2d(dim, dim * 2, 1, 1),
                nn.GELU(),
                nn.Conv2d(dim * 2, dim * 2, 1, 1),
                nn.GELU(),
                nn.Conv2d(dim * 2, gt_r**2, 1, 1),
            )
        else:
            self.pretrain_head = nn.Sequential(
                nn.Conv2d(dim, dim * 2, 1, 1),
                nn.GELU(),
                nn.Conv2d(dim * 2, dim * 2, 1, 1),
                nn.GELU(),
                nn.Conv2d(dim * 2, 81, 1, 1),
            )

        self.proj = nn.Conv2d(encoder_latent_dim, 256, 1)
        self.depth = decoder_depth
        self.decoder_layer = MemoryDecoderLayer(
            query_latent_dim=query_latent_dim,
            cost_latent_dim=cost_latent_dim,
            patch_size=patch_size,
            flow_or_pe=flow_or_pe,
            dropout=dropout,
            pe=pe,
            no_sc=no_sc,
        )

        if gma == "GMA":
            print("[Using GMA]")
            self.update_block = GMAUpdateBlock(
                r_16=r_16,
                cost_heads_num=cost_heads_num,
                query_latent_dim=query_latent_dim,
                hidden_dim=128,
            )
            self.att = Attention(dim=128, heads=1, max_pos_size=160, dim_head=128)
        elif gma == "GMA-SK":
            print("[Using GMA-SK]")
            self.update_block = SKUpdateBlock6_Deep_nopoolres_AllDecoder(
                cost_heads_num=cost_heads_num,
                query_latent_dim=query_latent_dim,
                hidden_dim=128,
            )
            self.att = Attention(dim=128, heads=1, max_pos_size=160, dim_head=128)
        else:
            print("[Not using GMA decoder]")
            self.update_block = BasicUpdateBlock(
                r_16=r_16,
                cost_heads_num=cost_heads_num,
                query_latent_dim=query_latent_dim,
                hidden_dim=128,
            )

        if r_16 > 0:
            print("[r_16 = {}]".format(r_16))

        if quater_refine:
            print("[Using Quater Refinement]")
            from .quater_upsampler import quater_upsampler

            self.quater_upsampler = quater_upsampler()

    def upsample_flow(self, flow, mask):
        """Upsample flow field [H/8, W/8, 2] -> [H, W, 2] using convex combination"""
        N, _, H, W = flow.shape
        mask = mask.view(N, 1, 9, 8, 8, H, W)
        mask = torch.softmax(mask, dim=2)

        up_flow = F.unfold(8 * flow, [3, 3], padding=1)
        up_flow = up_flow.view(N, 2, 9, 1, 1, H, W)

        up_flow = torch.sum(mask * up_flow, dim=2)
        up_flow = up_flow.permute(0, 1, 4, 2, 5, 3)
        return up_flow.reshape(N, 2, 8 * H, 8 * W)

    def sample_feature_map(self, coords, feat_t_quater, r=1):
        H, W = feat_t_quater.shape[-2:]
        xgrid, ygrid = coords.split([1, 1], dim=-1)
        xgrid = 2 * xgrid / (W - 1) - 1
        ygrid = 2 * ygrid / (H - 1) - 1

        grid = torch.cat([xgrid, ygrid], dim=-1)
        img = F.grid_sample(feat_t_quater, grid, align_corners=True)

        if mask:
            mask = (xgrid > -1) & (ygrid > -1) & (xgrid < 1) & (ygrid < 1)
            return img, mask.to(dtype=coords.dtype)

        return img

    def encode_flow_token(self, cost_maps, coords, r=4):
        """
        cost_maps   -   B*H1*W1, cost_heads_num, H2, W2
        coords      -   B, 2, H1, W1
        """
        coords = coords.permute(0, 2, 3, 1)
        batch, h1, w1, _ = coords.shape

        dx = torch.linspace(-r, r, 2 * r + 1, dtype=coords.dtype, device=coords.device)
        dy = torch.linspace(-r, r, 2 * r + 1, dtype=coords.dtype, device=coords.device)
        delta = torch.stack(torch.meshgrid(dy, dx, indexing="ij"), axis=-1)

        centroid = coords.reshape(batch * h1 * w1, 1, 1, 2)
        delta = delta.view(1, 2 * r + 1, 2 * r + 1, 2)
        coords = centroid + delta
        corr = bilinear_sampler(cost_maps, coords)

        corr = corr.view(batch, h1, w1, -1).permute(0, 3, 1, 2)
        return corr

    def forward(
        self,
        cost_memory,
        context,
        context_quater,
        feat_s_quater,
        feat_t_quater,
        data={},
        prev_flow=None,
        cost_patches=None,
    ):
        """
        memory: [B*H1*W1, H2'*W2', C]
        context: [B, D, H1, W1]
        """
        cost_maps = data["cost_maps"]
        coords0, coords1 = initialize_flow(context)

        if prev_flow is not None:
            forward_flow = forward_interpolate_batch(prev_flow)
            coords1 = coords1 + forward_flow

        # flow = coords1

        flow_predictions = []

        context = self.proj(context)
        net, inp = torch.split(context, [128, 128], dim=1)
        net = torch.tanh(net)
        inp = torch.relu(inp)
        if self.gma is not None:
            attention = self.att(inp)

        size = net.shape
        key, value = None, None

        for idx in range(self.depth):
            coords1 = coords1.detach()

            cost_forward = self.encode_flow_token(cost_maps, coords1)

            if self.use_patch:
                if self.detach_local:
                    _local_cost = self.encode_flow_token(
                        cost_patches, coords1 / 8.0, r=0
                    )
                    _local_cost = _local_cost.contiguous().detach()
                    query = self.flow_token_encoder(_local_cost)
                else:
                    query = self.flow_token_encoder(
                        self.encode_flow_token(cost_patches, coords1 / 8.0, r=0)
                    )
            else:
                if self.detach_local:
                    _local_cost = cost_forward.contiguous().detach()
                    query = self.flow_token_encoder(_local_cost)
                else:
                    query = self.flow_token_encoder(cost_forward)
            query = (
                query.permute(0, 2, 3, 1)
                .contiguous()
                .view(size[0] * size[2] * size[3], 1, self.dim)
            )

            if self.use_rpe:
                query_coord = coords1 - coords0
            else:
                query_coord = coords1
            cost_global, key, value = self.decoder_layer(
                query, key, value, cost_memory, query_coord, size, data["H3W3"]
            )

            if self.r_16 > 0:
                cost_forward_16 = self.encode_flow_token(
                    data["cost_maps_16"], coords1 * 2.0, r=(self.r_16 - 1) // 2
                )

                corr = torch.cat([cost_global, cost_forward, cost_forward_16], dim=1)
            else:
                corr = torch.cat([cost_global, cost_forward], dim=1)

            flow = coords1 - coords0

            if self.gma is not None:
                net, up_mask, delta_flow = self.update_block(
                    net, inp, corr, flow, attention
                )
            else:
                net, up_mask, delta_flow = self.update_block(net, inp, corr, flow)

            # flow = delta_flow
            coords1 = coords1 + delta_flow

            flow_up = self.upsample_flow(coords1 - coords0, up_mask)
            flow_predictions.append(flow_up)

        if self.quater_refine:
            coords1 = coords1.detach()
            new_size = context_quater.shape[-2:]
            flow = 2 * F.interpolate(
                coords1 - coords0, size=new_size, mode="bilinear", align_corners=True
            )
            flow_up = self.quater_upsampler(
                flow, context_quater, feat_s_quater, feat_t_quater, r=1
            )
            flow_predictions.append(flow_up)

        return flow_predictions, coords1 - coords0

    def pretrain_forward(
        self,
        cost_memory,
        context,
        data={},
        prev_flow=None,
        cost_patches=None,
        mask_for_patch1=None,
    ):
        cost_maps_outter = data["cost_maps_outter"].detach()
        cost_maps = data["cost_maps"].detach()

        _, _, H_outter, W_outter = cost_maps_outter.shape
        Bs, _, H_inner, W_inner = cost_maps.shape

        net, inp = torch.split(context, [128, 128], dim=1)

        size = net.shape
        B = size[0]
        key, value = None, None

        loss = 0

        if self.fix_pe:
            pad_l = pad_t = 0
            pad_r = (8 - W_inner % 8) % 8
            pad_b = (8 - H_inner % 8) % 8
            _H_inner = H_inner + pad_b
            _W_inner = W_inner + pad_r

            # number of keys for crossattentionlayer
            H_down = _H_inner // 8
            W_down = _W_inner // 8

            cost_maps = F.pad(cost_maps, (pad_l, pad_r, pad_t, pad_b))
            cost_maps_patches = F.unfold(cost_maps, kernel_size=8, padding=0, stride=8)
            mean = cost_maps_patches.mean(dim=1, keepdim=True)
            var = cost_maps_patches.var(dim=1, keepdim=True)
            cost_maps_patches = (cost_maps_patches - mean) / (var + 1.0e-6) ** 0.5
            cost_maps_patches = cost_maps_patches.reshape(Bs, 64, H_down, W_down)

            for idx_h, idx_w in zip(range(H_down), range(W_down)):
                query_coord = torch.zeros(
                    B, 2, H_inner, W_inner, device=cost_memory.device
                )
                query_coord[:, 0, :, :] = idx_w
                query_coord[:, 1, :, :] = idx_h
                cost_global, key, value = self.decoder_layer(
                    None,
                    key,
                    value,
                    cost_memory,
                    query_coord.detach(),
                    size,
                    data["H3W3"],
                )
                cost_forward_pred = self.pretrain_head(cost_global)

                target = (
                    cost_maps_patches[:, :, idx_h, idx_w]
                    .reshape(B, H_inner, W_inner, 64)
                    .permute(0, 3, 1, 2)
                )
                loss += ((cost_forward_pred - target) ** 2).mean()
        elif self.gt_r > 0:
            for idx in range(self.query_num):
                coords_outter = torch.rand(
                    B, 2, H_inner, W_inner, device=cost_memory.device
                )
                radius = (self.gt_r - 1) // 2
                if self.no_border:
                    coords_outter = (
                        torch.cat(
                            [
                                coords_outter[:, 0:1, :, :] * (W_outter - self.gt_r),
                                coords_outter[:, 1:, :, :] * (H_outter - self.gt_r),
                            ],
                            dim=1,
                        )
                        + radius
                    )
                else:
                    coords_outter = torch.cat(
                        [
                            coords_outter[:, 0:1, :, :] * W_outter,
                            coords_outter[:, 1:, :, :] * H_outter,
                        ],
                        dim=1,
                    )

                coords_inner = torch.cat(
                    [
                        coords_outter[:, 0:1, :, :] - self.W_offset // 8,
                        coords_outter[:, 1:, :, :] - self.H_offset // 8,
                    ],
                    dim=1,
                )

                cost_forward_outter = self.encode_flow_token(
                    cost_maps_outter.detach(), coords_outter.detach(), r=radius
                )

                query_coord = coords_inner
                cost_forward = self.encode_flow_token(cost_maps.detach(), coords_inner)
                query = self.flow_token_encoder(cost_forward)
                query = (
                    query.permute(0, 2, 3, 1)
                    .contiguous()
                    .view(size[0] * size[2] * size[3], 1, self.dim)
                )
                cost_global, key, value = self.decoder_layer(
                    query,
                    key,
                    value,
                    cost_memory,
                    query_coord.detach(),
                    size,
                    data["H3W3"],
                )

                cost_forward_pred = self.pretrain_head(cost_global)
                mean = cost_forward_outter.mean(dim=1, keepdim=True)
                var = cost_forward_outter.var(dim=1, keepdim=True)
                cost_forward_outter = (cost_forward_outter - mean) / (
                    var + 1.0e-6
                ) ** 0.5

                loss += ((cost_forward_pred - cost_forward_outter) ** 2).mean()
        else:
            for idx in range(self.query_num):
                coords_outter = torch.rand(
                    B, 2, H_inner, W_inner, device=cost_memory.device
                )
                if self.no_border:
                    coords_outter = (
                        torch.cat(
                            [
                                coords_outter[:, 0:1, :, :] * (W_outter - 8),
                                coords_outter[:, 1:, :, :] * (H_outter - 8),
                            ],
                            dim=1,
                        )
                        + 4.0
                    )
                else:
                    coords_outter = torch.cat(
                        [
                            coords_outter[:, 0:1, :, :] * W_outter,
                            coords_outter[:, 1:, :, :] * H_outter,
                        ],
                        dim=1,
                    )

                coords_inner = torch.cat(
                    [
                        coords_outter[:, 0:1, :, :] - self.W_offset // 8,
                        coords_outter[:, 1:, :, :] - self.H_offset // 8,
                    ],
                    dim=1,
                )

                cost_forward_outter = self.encode_flow_token(
                    cost_maps_outter.detach(), coords_outter.detach()
                )

                query_coord = coords_inner
                cost_global, key, value = self.decoder_layer(
                    None,
                    key,
                    value,
                    cost_memory,
                    query_coord.detach(),
                    size,
                    data["H3W3"],
                )

                cost_forward_pred = self.pretrain_head(cost_global)
                mean = cost_forward_outter.mean(dim=1, keepdim=True)
                var = cost_forward_outter.var(dim=1, keepdim=True)
                cost_forward_outter = (cost_forward_outter - mean) / (
                    var + 1.0e-6
                ) ** 0.5

                loss += ((cost_forward_pred - cost_forward_outter) ** 2).mean()

        return loss



""" AvgPool2d w/ Same Padding

Hacked together by / Copyright 2020 Ross Wightman
"""

import torch.nn as nn
import torch.nn.functional as F
from typing import List

from .layer_helpers import to_2tuple
from .padding import pad_same, get_padding_value


def avg_pool2d_same(
    x,
    kernel_size: List[int],
    stride: List[int],
    padding: List[int] = (0, 0),
    ceil_mode: bool = False,
    count_include_pad: bool = True,
):
    # FIXME how to deal with count_include_pad vs not for external padding?
    x = pad_same(x, kernel_size, stride)
    return F.avg_pool2d(x, kernel_size, stride, (0, 0), ceil_mode, count_include_pad)


class AvgPool2dSame(nn.AvgPool2d):
    """Tensorflow like 'SAME' wrapper for 2D average pooling"""

    def __init__(
        self,
        kernel_size: int,
        stride=None,
        padding=0,
        ceil_mode=False,
        count_include_pad=True,
    ):
        kernel_size = to_2tuple(kernel_size)
        stride = to_2tuple(stride)
        super(AvgPool2dSame, self).__init__(
            kernel_size, stride, (0, 0), ceil_mode, count_include_pad
        )

    def forward(self, x):
        x = pad_same(x, self.kernel_size, self.stride)
        return F.avg_pool2d(
            x,
            self.kernel_size,
            self.stride,
            self.padding,
            self.ceil_mode,
            self.count_include_pad,
        )


def max_pool2d_same(
    x,
    kernel_size: List[int],
    stride: List[int],
    padding: List[int] = (0, 0),
    dilation: List[int] = (1, 1),
    ceil_mode: bool = False,
):
    x = pad_same(x, kernel_size, stride, value=-float("inf"))
    return F.max_pool2d(x, kernel_size, stride, (0, 0), dilation, ceil_mode)


class MaxPool2dSame(nn.MaxPool2d):
    """Tensorflow like 'SAME' wrapper for 2D max pooling"""

    def __init__(
        self, kernel_size: int, stride=None, padding=0, dilation=1, ceil_mode=False
    ):
        kernel_size = to_2tuple(kernel_size)
        stride = to_2tuple(stride)
        dilation = to_2tuple(dilation)
        super(MaxPool2dSame, self).__init__(
            kernel_size, stride, (0, 0), dilation, ceil_mode
        )

    def forward(self, x):
        x = pad_same(x, self.kernel_size, self.stride, value=-float("inf"))
        return F.max_pool2d(
            x, self.kernel_size, self.stride, (0, 0), self.dilation, self.ceil_mode
        )


def create_pool2d(pool_type, kernel_size, stride=None, **kwargs):
    stride = stride or kernel_size
    padding = kwargs.pop("padding", "")
    padding, is_dynamic = get_padding_value(
        padding, kernel_size, stride=stride, **kwargs
    )
    if is_dynamic:
        if pool_type == "avg":
            return AvgPool2dSame(kernel_size, stride=stride, **kwargs)
        elif pool_type == "max":
            return MaxPool2dSame(kernel_size, stride=stride, **kwargs)
        else:
            assert False, f"Unsupported pool type {pool_type}"
    else:
        if pool_type == "avg":
            return nn.AvgPool2d(kernel_size, stride=stride, padding=padding, **kwargs)
        elif pool_type == "max":
            return nn.MaxPool2d(kernel_size, stride=stride, padding=padding, **kwargs)
        else:
            assert False, f"Unsupported pool type {pool_type}"



from torch import nn
from functools import partial
import numpy as np


class PreNormResidual(nn.Module):
    def __init__(self, dim, fn):
        super().__init__()
        self.fn = fn
        self.norm = nn.LayerNorm(dim)

    def forward(self, x):
        return self.fn(self.norm(x)) + x


def FeedForward(dim, expansion_factor=4, dropout=0.0, dense=nn.Linear):
    return nn.Sequential(
        dense(dim, dim * expansion_factor),
        nn.GELU(),
        nn.Dropout(dropout),
        dense(dim * expansion_factor, dim),
        nn.Dropout(dropout),
    )


class MLPMixerLayer(nn.Module):
    def __init__(self, dim, cost_latent_token_num, mlp_expansion_factor, dropout=0.0):
        super(MLPMixerLayer, self).__init__()

        # print(f"use mlp mixer layer")
        K = cost_latent_token_num
        expansion_factor = mlp_expansion_factor
        chan_first, chan_last = partial(nn.Conv1d, kernel_size=1), nn.Linear

        self.mlpmixer = nn.Sequential(
            PreNormResidual(dim, FeedForward(K, expansion_factor, dropout, chan_first)),
            PreNormResidual(
                dim, FeedForward(dim, expansion_factor, dropout, chan_last)
            ),
        )

    def compute_params(self):
        num = 0
        for param in self.mlpmixer.parameters():
            num += np.prod(param.size())

        return num

    def forward(self, x):
        """
        x: [BH1W1, K, D]
        """

        return self.mlpmixer(x)



from .ccmr import *
from .craft import *
from .csflow import *
from .dicl import *
from .dip import *
from .dpflow import *
from .fastflownet import *
from .flow1d import *
from .flowformer import *
from .flowformerplusplus import *
from .flownet import *
from .gma import *
from .gmflow import *
from .gmflownet import *
from .hd3 import *
from .irr.pwcnet import *
from .lcv import *
from .liteflownet import *
from .llaflow import *
from .maskflownet import *
from .matchflow import *
from .memflow import *
from .ms_raft_plus import *
from .neuflow import *
from .neuflow2 import *
from .pwcnet import *
from .raft import *
from .rapidflow import *
from .rpknet import *
from .sea_raft import *
from .scopeflow import *
from .scv.scv import *
from .separableflow import *
from .skflow import *
from .splatflow import *
from .starflow import *
from .streamflow import *
from .unimatch import *
from .vcn import *
from .videoflow import *



import torch
from torch import nn, einsum
from einops import rearrange


# max_pos_size = 160
class RelPosEmb(nn.Module):
    def __init__(self, max_pos_size, dim_head):
        super().__init__()
        self.rel_height = nn.Embedding(2 * max_pos_size - 1, dim_head)
        self.rel_width = nn.Embedding(2 * max_pos_size - 1, dim_head)

        deltas = torch.arange(max_pos_size).view(1, -1) - torch.arange(
            max_pos_size
        ).view(-1, 1)
        # rel_ind[i, j] = j - i + 159.
        rel_ind = deltas + max_pos_size - 1
        self.register_buffer("rel_ind", rel_ind)

    def forward(self, q):
        # q: [8, 1, 46, 62, 128]
        batch, heads, h, w, c = q.shape
        # self.rel_ind[:h, :h]: [46, 46]
        # self.rel_ind[:w, :w]: [62, 62]
        # rel_ind[i,j] = j - i + 159, precomputed distance between i, j.
        # This assumes the input x (from which q is derived) is precisely on the grid.
        # This is fine when we do self-attention on x.
        # However, it will be somewhat limiting if we use RelPosEmb on cross-attention between two frames,
        # particularly when we use flow_init != 0 (on sintel),
        # we better get the positional encodings of x according to flow_init, instead of the grid of x.
        # However, an accurate computation of the relative distances between all input units is expensive.
        # Since values in flow_init are usually small, this inaccuracy may be negligible.
        height_emb = self.rel_height(self.rel_ind[:h, :h].reshape(-1))
        width_emb = self.rel_width(self.rel_ind[:w, :w].reshape(-1))

        # height_emb: [46*46, 128] => [46, 46, 1,  128]
        # width_emb:  [62*62, 128] => [62, 1,  62, 128]
        # height_emb[i, j]: the embedding of element at (i,j) as a function of the height difference (i-j).
        # width_emb[i, j]:  the embedding of element at (i,j) as a function of the width  difference (i-j).
        height_emb = rearrange(height_emb, "(x u) d -> x u () d", x=h)
        width_emb = rearrange(width_emb, "(y v) d -> y () v d", y=w)

        # outer product? y, uv -> y u v              b  h  x   y   d        x  u   v   d
        # height_score: [8, 1, 46, 62, 46, 1]    <= [8, 1, 46, 62, 128] * [46, 46, 1,  128]
        # width_score:  [8, 1, 46, 62, 1,  62]
        height_score = einsum("b h x y d, x u v d -> b h x y u v", q, height_emb)
        width_score = einsum("b h x y d, y u v d -> b h x y u v", q, width_emb)
        # height_score + width_score: [8, 1, 46, 62, 46, 62], 65071232 elements.
        return height_score + width_score


class Attention(nn.Module):
    def __init__(
        self,
        *,
        dim,
        max_pos_size=100,
        heads=4,
        dim_head=128,
    ):
        super().__init__()
        self.heads = heads
        self.scale = dim_head**-0.5
        inner_dim = heads * dim_head

        self.to_qk = nn.Conv2d(dim, inner_dim * 2, 1, bias=False)

        self.pos_emb = RelPosEmb(max_pos_size, dim_head)
        self.pos_embed_weight = 1.0

    def forward(self, fmap):
        heads, b, c, h, w = self.heads, *fmap.shape

        # q, k: [8, 128, 46, 62]
        q, k = self.to_qk(fmap).chunk(2, dim=1)

        # q, k: [8, 1, 46, 62, 128]
        q, k = map(lambda t: rearrange(t, "b (h d) x y -> b h x y d", h=heads), (q, k))
        # Why not scale k?
        q = self.scale * q

        if self.args.position_only:
            sim = self.pos_emb(q)

        elif self.args.position_and_content:
            # [..., 46, 62, ...] . [..., 46, 62, ...] => [..., 46, 62, 46, 62]
            sim_content = einsum("b h x y d, b h u v d -> b h x y u v", q, k)
            sim_pos = self.pos_emb(q)
            sim = sim_content + self.pos_embed_weight * sim_pos

        else:
            # q, k: [B, 1, 46, 62, 128]
            # sim:  [B, 1, 46, 62, 46, 62]
            sim = einsum("b h x y d, b h u v d -> b h x y u v", q, k)

        sim = rearrange(sim, "b h x y u v -> b h (x y) (u v)")
        attn = sim.softmax(dim=-1)

        return attn


# Aggregate output is dim-dimensional, same as the input. No FFN is used.
class Aggregate(nn.Module):
    def __init__(
        self,
        dim,
        heads=4,
        dim_head=128,
    ):
        super().__init__()
        self.heads = heads
        self.scale = dim_head**-0.5
        inner_dim = heads * dim_head

        self.to_v = nn.Conv2d(dim, inner_dim, 1, bias=False)

        self.gamma = nn.Parameter(torch.zeros(1))

        if dim != inner_dim:
            self.project = nn.Conv2d(inner_dim, dim, 1, bias=False)
        else:
            self.project = None

    def forward(self, attn, fmap):
        heads, b, c, h, w = self.heads, *fmap.shape

        v = self.to_v(fmap)
        v = rearrange(v, "b (h d) x y -> b h (x y) d", h=heads)
        out = einsum("b h i j, b h j d -> b h i d", attn, v)
        out = rearrange(out, "b h (x y) d -> b (h d) x y", x=h, y=w)

        # project is None for GMA.
        if self.project is not None:
            out = self.project(out)

        out = fmap + self.gamma * out

        return out


if __name__ == "__main__":
    att = Attention(dim=128, heads=1)
    fmap = torch.randn(2, 128, 40, 90)
    out = att(fmap)

    print(out.shape)
